from typing import Optional

import SimpleITK as sitk


def read_image(path: str, series_id: Optional[str] = None) -> sitk.Image:
    reader = sitk.ImageSeriesReader()
    dicom_names = reader.GetGDCMSeriesFileNames(
        path, seriesID=series_id if series_id else ""
    )
    reader.SetFileNames(dicom_names)
    reader.MetaDataDictionaryArrayUpdateOn()
    reader.LoadPrivateTagsOn()

    return reader.Execute()



import os
from pathlib import Path

import structlog

from imgtools.logging.logging_config import DEFAULT_LOG_LEVEL, LoggingManager

# Set the default log level from the environment variable or use the default
DEFAULT_OR_ENV = os.environ.get("IMGTOOLS_LOG_LEVEL", DEFAULT_LOG_LEVEL)

LOG_DIR_NAME = Path(".imgtools/logs")
DEFAULT_LOG_FILENAME = "imgtools.log"


def get_logger(name: str, level: str = "INFO") -> structlog.stdlib.BoundLogger:
    """
    Retrieve a logger with the specified log level.

    Parameters
    ----------
    name : str
                    Name of Logger Instance
    level : str
        Desired logging level.

    Returns
    -------
    logging.Logger
        Configured logger instance.
    """
    logging_manager = LoggingManager(name)
    env_level = logging_manager.env_level

    if env_level not in (level, DEFAULT_LOG_LEVEL):
        logging_manager.get_logger().warning(
            f"Environment variable {name}_LOG_LEVEL is {env_level} "
            f"but you are setting it to {level}"
        )
    return logging_manager.configure_logging(level=level)


logger = get_logger("imgtools", DEFAULT_OR_ENV)


if __name__ == "__main__":
    # Example usage of the logging manager
    logger.debug("This is a debug message")
    logger.info("This is an info message")
    logger.warning("This is a warning message")
    logger.error("This is an error message")
    logger.critical("This is a critical message")

    readii_logger = get_logger("readii", "DEBUG")
    readii_logger.debug("This is a debug message")
    readii_logger.info("This is an info message")
    readii_logger.warning("This is a warning message")
    readii_logger.error("This is an error message")



import logging.config
import os
from pathlib import Path
from typing import Dict, List

import structlog
from structlog.processors import CallsiteParameter, CallsiteParameterAdder
from structlog.typing import Processor

from imgtools.logging.processors import (
    CallPrettifier,
    ESTTimeStamper,
    PathPrettifier,
)

VALID_LOG_LEVELS = {"DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"}
DEFAULT_LOG_LEVEL = "WARNING"


class LoggingManager:
    """
    Manages the configuration and initialization of a structured logger.

    This class provides flexible options for configuring log levels, formats,
    and output destinations.

    Parameters
    ----------
    name : str
        Name of the logger instance.

    Attributes
    ----------
    name : str
        Name of the logger instance.
    base_dir : Path
        Base directory for path prettification.
                    i.e if base_dir is /home/user/project and logs include a path like /home/user/project/data/file.txt
                    then the path will be prettified to data/file.txt
    level : str
        Log level for the logger.


    Methods
    -------
    get_logger()
        Retrieve the configured logger instance.
    configure_logging(level=None)
        Dynamically adjust logging settings.

    Raises
    ------
    ValueError
        If an invalid log level is provided.
    RuntimeError
        If the log directory or file cannot be created.

    Examples
    --------
    Initialize with default settings:
        >>> manager = LoggingManager(name="mypackage")
        >>> logger = manager.get_logger()
        >>> logger.info("Info message")
    """

    def __init__(
        self,
        name: str,
        base_dir: Path | None = None,
    ) -> None:
        self.name = name
        self.base_dir = base_dir or Path.cwd()
        self.level = self.env_level
        self._initialize_logger()

    @property
    def env_level(self) -> str:
        return os.environ.get(
            f"{self.name}_LOG_LEVEL".upper(), DEFAULT_LOG_LEVEL
        ).upper()

    @property
    def base_logging_config(self) -> Dict:
        """
        Create the basic logging configuration settings.

        Returns
        -------
        dict
            Base logging configuration.
        """
        return {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "console": {
                    "()": structlog.stdlib.ProcessorFormatter,
                    "processors": [
                        CallPrettifier(concise=True),
                        structlog.stdlib.ProcessorFormatter.remove_processors_meta,
                        structlog.dev.ConsoleRenderer(
                            exception_formatter=structlog.dev.RichTracebackFormatter(
                                width=-1, show_locals=False
                            ),
                        ),
                    ],
                    "foreign_pre_chain": self.pre_chain,
                },
            },
            "handlers": {
                "console": {
                    "class": "logging.StreamHandler",
                    "formatter": "console",
                },
            },
            "loggers": {
                self.name: {
                    "handlers": ["console"],
                    "level": self.level,
                    "propagate": False,
                },
            },
        }

    @property
    def pre_chain(self) -> List[Processor]:
        return [
            structlog.stdlib.add_log_level,
            ESTTimeStamper(),
            structlog.stdlib.add_logger_name,
            structlog.stdlib.PositionalArgumentsFormatter(),
            CallsiteParameterAdder(
                [
                    CallsiteParameter.MODULE,
                    CallsiteParameter.FUNC_NAME,
                    CallsiteParameter.LINENO,
                ]
            ),
            PathPrettifier(base_dir=self.base_dir),
            structlog.stdlib.ExtraAdder(),
            structlog.processors.StackInfoRenderer(),
        ]

    def _initialize_logger(self) -> None:
        """
        Initialize the logger with the current configuration.
        """

        logging.config.dictConfig(self.base_logging_config)
        structlog.configure(
            processors=[
                *self.pre_chain,
                structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
            ],
            logger_factory=structlog.stdlib.LoggerFactory(),
            wrapper_class=structlog.stdlib.BoundLogger,
            cache_logger_on_first_use=True,
        )

    def get_logger(self) -> structlog.stdlib.BoundLogger:
        """
        Retrieve the logger instance.

        Returns
        -------
        structlog.stdlib.BoundLogger
            Configured logger instance.
        """
        return structlog.get_logger(self.name)

    def configure_logging(
        self, level: str = DEFAULT_LOG_LEVEL
    ) -> structlog.stdlib.BoundLogger:
        """
        Dynamically adjust logging settings.

        Parameters
        ----------
        level : str, optional
                        Set the log level.

        Returns
        -------
        structlog.stdlib.BoundLogger
                        Updated logger instance.

        Raises
        ------
        ValueError
                        If an invalid log level is specified.
        """
        level_upper = level.upper()
        if level_upper not in VALID_LOG_LEVELS:
            msg = f"Invalid logging level: {level}"
            raise ValueError(msg)

        # Store the old level for logging the change
        old_level = self.level
        self.level = level_upper

        self._initialize_logger()
        logger = self.get_logger()

        # Log the level change
        if old_level != self.level:
            logger.info(
                "Log level changed", old_level=old_level, new_level=self.level
            )

        return logger



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Sequence

import numpy as np
import SimpleITK as sitk

if TYPE_CHECKING:
    from .direction import Direction
    from .helper_types import Coordinate3D, Size3D, Spacing3D


@dataclass(frozen=True)
class ImageGeometry:
    """Represents the geometry of a 3D image."""

    size: Size3D
    origin: Coordinate3D
    direction: Direction
    spacing: Spacing3D


# Type alias for Image or Array
ImageOrArray = sitk.Image | np.ndarray


class MedImage:
    def __init__(
        self,
        image: ImageOrArray,
        origin: Sequence[float] | None = None,
        direction: Sequence[float] | None = None,
        spacing: Sequence[float] | None = None,
    ) -> None:
        pass


# The following goals and code was written by
# the original author of the codebase, considering it as a possible implementation

# Goals for this module:
# - One Image Class to Rule Them All (no more juggling numpy & sitk)
# - consistent indexing (z, y, x) (maybe fancy indexing?)
# - can be created either from array or sitk.Image
# - implements all basic operators of sitk.Image
# - easy conversion to array or sitk
# - method to apply arbitrary sitk filter
# - nicer repr

# TODO:
# - better support for masks (e.g. what happens when we pass a one-hot encoded mask?)
# - (optional) specify indexing order when creating new Image (sitk or numpy)


# class Image:
#     def __init__(
#         self,
#         image: ImageOrArray,
#         geometry: ImageGeometry | None = None,
#         origin: Sequence[float] | None = None,
#         direction: Sequence[float] | None = None,
#         spacing: Sequence[float] | None = None,
#     ) -> None:
#         if isinstance(image, sitk.Image):
#             self._image = image
#         elif isinstance(image, np.ndarray):
#             if geometry is None and any(
#                 (origin is None, direction is None, spacing is None)
#             ):
#                 raise ValueError(
#                     "If image is a Numpy array, either geometry must be specified."
#                 )

#             if geometry is not None:
#                 _, origin, direction, spacing = geometry

#             self._image = sitk.GetImageFromArray(image)
#             self._image.SetOrigin(origin[::-1])
#             direction = tuple(direction)
#             self._image.SetDirection(
#                 direction[:-3] + direction[3:6] + direction[:3]
#             )
#             self._image.SetSpacing(spacing[::-1])
#         else:
#             msg = f"image must be either numpy.ndarray or SimpleITK.Image, not {type(image)}."
#             raise TypeError(msg)

#     @property
#     def size(self):
#         return self._image.GetSize()[::-1]

#     @property
#     def origin(self):
#         return self._image.GetOrigin()[::-1]

#     @property
#     def direction(self):
#         direction = self._image.GetDirection()
#         direction = direction[:-3] + direction[3:6] + direction[:3]
#         return direction

#     @property
#     def spacing(self):
#         return self._image.GetSpacing()[::-1]

#     @property
#     def geometry(self):
#         return ImageGeometry(
#             size=self.size,
#             origin=self.origin,
#             direction=self.direction,
#             spacing=self.spacing,
#         )

#     @property
#     def ndim(self):
#         return len(self.size)

#     @property
#     def dtype(self):
#         return self._image.GetPixelIDType()

#     def astype(self, new_type):
#         return Image(sitk.Cast(self._image, new_type))

#     def to_sitk_image(self):
#         return self._image

#     def to_numpy(
#         self, return_geometry: bool = False, view: bool = False
#     ) -> Union[np.ndarray, tuple[np.ndarray, ImageGeometry]]:
#         if view:
#             array = sitk.GetArrayViewFromImage(self._image)
#         else:
#             array = sitk.GetArrayFromImage(self._image)
#         if return_geometry:
#             return array, self.geometry
#         return array

#     def __getitem__(self, idx: int | slice) -> Union[Image, Any]:  # noqa
#         if isinstance(idx, (int, slice)):
#             idx = (idx,)
#         if len(idx) < self.ndim:
#             idx += (slice(None),) * (self.ndim - len(idx))

#         idx = idx[::-1]  # SimpleITK uses (x, y, z) ordering internally

#         value = self._image[idx]

#         try:  # XXX there probably is a nicer way to check if value is a scalar
#             return Image(value)
#         except TypeError:
#             return value

#     def __setitem__(self, idx: int | slice, value: Any) -> None:  # noqa
#         if isinstance(idx, (int, slice)):
#             idx = (idx,)
#         if len(idx) < self.ndim:
#             idx += (slice(None),) * (self.ndim - len(idx))

#         idx = idx[::-1]  # SimpleITK uses (x, y, z) ordering internally

#         value = self._image[idx]

#         try:  # XXX there probably is a nicer way to check if value is a scalar
#             return Image(value)
#         except TypeError:
#             return value

#     def apply_filter(self, sitk_filter: sitk.ImageFilter) -> Union[Image, Any]:  # noqa
#         result = sitk_filter.Execute(self._image)
#         if isinstance(result, sitk.Image):
#             return Image(result)
#         else:
#             return result

#     def __neg__(self) -> Image:
#         return Image(-self._image)

#     def __abs__(self) -> Image:
#         return Image(abs(self._image))

#     def __invert__(self) -> Image:
#         return Image(~self._image)

#     def __add__(self, other: Union[Image, Image]) -> Image:
#         other_val = getattr(other, "_image", other)
#         return Image(self._image + other_val)

#     def __sub__(self, other: Union[Image, Image]) -> Image:
#         other_val = getattr(other, "_image", other)
#         return Image(self._image - other_val)

#     def __mul__(self, other: Union[Image, Image]) -> Image:
#         other_val = getattr(other, "_image", other)
#         return Image(self._image * other_val)

#     def __div__(self, other: Union[Image, Image]) -> Image:
#         other_val = getattr(other, "_image", other)
#         return Image(self._image / other_val)

#     def __floordiv__(self, other: Union[Image, Image]) -> Image:
#         other_val = getattr(other, "_image", other)
#         return Image(self._image // other_val)

#     def __pow__(self, other: Union[Image, Image]) -> Image:
#         other_val = getattr(other, "_image", other)
#         return Image(self._image**other_val)

#     def __iadd__(self, other: Union[Image, Image]) -> Image:
#         other_val = getattr(other, "_image", other)
#         self._image += other_val

#     def __isub__(self, other: Union[Image, Image]) -> Image:
#         other_val = getattr(other, "_image", other)
#         self._image -= other_val

#     def __imul__(self, other: Union[Image, Image]) -> Image:
#         other_val = getattr(other, "_image", other)
#         self._image *= other_val

#     def __idiv__(self, other: Union[Image, Image]) -> Image:
#         other_val = getattr(other, "_image", other)
#         self._image /= other_val

#     def __ifloordiv__(self, other: Union[Image, Image]) -> Image:
#         other_val = getattr(other, "_image", other)
#         self._image //= other_val

#     def __iter__(self) -> Iterator:
#         pass

#     def __repr__(self) -> str:
#         return (
#             f"Image(image={self._image}, origin={self.origin}, "
#             f"spacing={self.spacing}, direction={self.direction})"
#         )

#     def __str__(self) -> str:
#         return (
#             f"origin = {self.origin}\n"
#             f"spacing = {self.spacing}\n"
#             f"direction = {self.direction}\n"
#             f"values = \n{self.to_numpy(view=True)}"
#         )



from pathlib import Path
import pytest
from rich import print as rprint
import pandas as pd
from imgtools.crawler import crawl

"""
This stuff will be hard coded unfortunately, and will be tedious to adjust if
we ever make changes to the test data. but its the best we can do for now.

idea test would be 

'collection' : {
  'modality' : dict of counts,
  'references' : make sure all the references collected stay the same so we can refactor without breaking tests
}

"""


COLLECTIONS = {
	"4D-Lung": {
		"CT": 70,
		"RTSTRUCT": 50,
	},
	"Adrenal-ACC-Ki67-Seg": {
		"CT": 6,
		"SEG": 2,
	},
	"CC-Tumor-Heterogeneity": {
		"CT": 6,
		"PT": 6,
		"MR": 48,
		"RTSTRUCT": 6,
		"REG": 33,
	},
	"LIDC-IDRI": {
		"CT": 3,
		"SEG": 15,
		"SR": 15,
		"CR": 1,
	},
	"Mediastinal-Lymph-Node-SEG": {
		"CT": 2,
		"SEG": 2,
	},
	"NSCLC-Radiomics": {
		"CT": 2,
		"RTSTRUCT": 2,
		"SEG": 2,
	},
	"NSCLC_Radiogenomics": {
		"PT": 6,
		"CT": 6,
		"SEG": 3,
	},
	"Prostate-Anatomical-Edge-Cases": {
		"CT": 2,
		"RTSTRUCT": 2,
	},
	"QIN-PROSTATE-Repeatability": {
		"MR": 20,
		"SEG": 12,
		"SR": 12,
	},
	"Soft-tissue-Sarcoma": {
		"MR": 8,
		"CT": 2,
		"PT": 2,
		"RTSTRUCT": 12,
	},
	"Vestibular-Schwannoma-SEG": {
		"MR": 4,
		"RTPLAN": 4,
		"RTSTRUCT": 4,
		"RTDOSE": 4,
	},
  "Head-Neck-PET-CT": {
    "CT": 2,
    "PT": 2,
    "RTSTRUCT": 4,
    "RTPLAN": 2,
    "RTDOSE": 2,
  }
}


@pytest.mark.parametrize("dataset_key", [
  "Adrenal-ACC-Ki67-Seg",
  "LIDC-IDRI",
  "Mediastinal-Lymph-Node-SEG",
  # "NSCLC-Radiomics",
  # "NSCLC_Radiogenomics",
  "Prostate-Anatomical-Edge-Cases",
  # "QIN-PROSTATE-Repeatability",
  "Soft-tissue-Sarcoma",
  "Vestibular-Schwannoma-SEG",
  "Head-Neck-PET-CT"
])
def test_shared_datadir_datasets(data_paths, dataset_key) -> None:
  """
  Test that each dataset key exists in the copied shared dataset directory.
  """
  rprint(f"Checking dataset: {dataset_key}")
  assert dataset_key in data_paths, f"Dataset key {dataset_key} not found"
  assert data_paths[dataset_key].exists(), f"Dataset {dataset_key} not found"


  db = crawl(
    top=data_paths[dataset_key],
    n_jobs=1,
  )

  assert isinstance(db, dict), "Expected a dictionary output"
  assert len(db) > 0, "Expected non-empty dictionary output"

  top_dir = Path(data_paths[dataset_key])
  imgtools_dir = top_dir.parent / ".imgtools"
  crawl_path = imgtools_dir / f"imgtools_{dataset_key}.csv"

  assert crawl_path.exists(), f"Expected crawl file {crawl_path} not found"
  assert crawl_path.with_suffix(".json").exists()  , f"Expected crawl file {crawl_path.with_suffix('.json')} not found"

  # read in the csv, get the "modality" column, use collections counter to count the unique values
  # verify that the counts match the expected counts within the COLLECTIONS dictionary

  df = pd.read_csv(crawl_path, index_col=0, dtype={"modality": str})
  modality_counts = df["modality"].value_counts()

  for modality, count in modality_counts.items():
    assert modality in COLLECTIONS[dataset_key], f"Modality {modality} not found in COLLECTIONS"
    assert count == COLLECTIONS[dataset_key][modality], f"Expected {COLLECTIONS[dataset_key][modality]} {modality} files, found {count}"




import os
import pathlib
from pathlib import Path
from typing import Generator, Tuple
from urllib import request
from zipfile import ZipFile
import json
import pytest
from filelock import FileLock


from imgtools.datasets.github_helper import (
    GitHubRelease,
    MedImageTestData,
)
from imgtools.logging import logger  # type: ignore

# from .conf_helpers import ensure_data_dir_exists  # type: ignore


@pytest.fixture(scope="session")
def data_dir() -> pathlib.Path:
    data_dir = Path(__file__).parent.parent / "data"
    if not data_dir.exists():
        data_dir.mkdir(parents=True, exist_ok=True)
    return data_dir


@pytest.fixture(scope="session")
def download_all_test_data(data_dir: pathlib.Path, worker_id: str | None) -> dict[str, Path]:  

    # idk how this works like in the pytest-xdist docs
    # leaving in case we need to use it later
    # if worker_id == "master":
    #     # following the pytest-xdist docs, we only want to download the data once

    #     manager = MedImageTestData()
    #     latest_release: GitHubRelease = manager.get_latest_release()
    #     # banned = ["4D-Lungs", "CC-Tumor-Heterogeneisty.tar"]
    #     banned = ["4D-Lung", "CC-Tumor-Heterogeneity.tar"]
    #     selected_tests = list(
    #         filter(
    #             lambda asset: not any(
    #                 asset.name.startswith(nope) for nope in banned
    #             ),
    #             latest_release.assets,
    #         )
    #     )
    #     extracted_paths = manager.download(data_dir, assets=selected_tests)
    #     dataset_path_mapping = {
    #         unzip_path.name: unzip_path for unzip_path in extracted_paths
    #     }
    #     return dataset_path_mapping

    lock_path = data_dir / "download_all_test_data"

    with FileLock(str(lock_path) + ".lock"):
        if lock_path.with_suffix(".json").exists():
            # Load existing dataset mapping if already downloaded
            logger.info("Loading pre-downloaded test dataset metadata...")
            loaded_json =  json.loads(lock_path.with_suffix(".json").read_text())
            dataset_path_mapping = {k: Path(v) for k, v in loaded_json.items()}
            return dataset_path_mapping
        # Download QC dataset
        logger.info("Downloading the test dataset...")
        manager = MedImageTestData()
        latest_release: GitHubRelease = manager.get_latest_release()
        banned = ["4D-Lung", "CC-Tumor-Heterogeneity.tar"]
        selected_tests = list(
            filter(
                lambda asset: not any(
                    asset.name.startswith(nope) for nope in banned
                ),
                latest_release.assets,
            )
        )
        extracted_paths = manager.download(data_dir, assets=selected_tests)
        dataset_path_mapping = {
            unzip_path.name: unzip_path for unzip_path in extracted_paths
        }
        # Save dataset mapping to avoid re-downloading in future runs
        lock_path.with_suffix(".json").write_text(
            json.dumps({k: str(v) for k, v in dataset_path_mapping.items()})
        )
        return dataset_path_mapping

@pytest.fixture(scope="session")
def download_old_test_data(data_dir: pathlib.Path) -> dict[str, Path]:
    """
    We have a few old tests that we want to keep around for now.
    this is mainly the quebec dataset
    """

    lock_path = data_dir / "download_old_test_data"

    with FileLock(str(lock_path) + ".lock"):
        if lock_path.with_suffix(".json").exists():
            # Load existing dataset mapping if already downloaded
            logger.info("Loading pre-downloaded old test dataset metadata...")
            loaded_json = json.loads(lock_path.with_suffix(".json").read_text())
            dataset_path_mapping = {k: Path(v) for k, v in loaded_json.items()}
            return dataset_path_mapping

        # Download QC dataset
        logger.info("Downloading the old test dataset...")
        quebec_data_path = data_dir / "Head-Neck-PET-CT"

        if not quebec_data_path.exists():
            quebec_data_url = "https://github.com/bhklab/tcia_samples/blob/main/Head-Neck-PET-CT.zip?raw=true"
            quebec_zip_path = quebec_data_path.with_suffix(".zip")

            quebec_data_path.mkdir(parents=True, exist_ok=True)
            request.urlretrieve(quebec_data_url, quebec_zip_path)
            with ZipFile(quebec_zip_path, "r") as zipfile:
                zipfile.extractall(quebec_data_path)
            quebec_zip_path.unlink()

        assert quebec_data_path.exists(), f"Quebec data not found at {quebec_data_path}"
        dataset_path_mapping = {
            "Head-Neck-PET-CT": quebec_data_path,
        }

        # Save dataset mapping to avoid re-downloading in future runs
        lock_path.with_suffix(".json").write_text(
            json.dumps({k: str(v) for k, v in dataset_path_mapping.items()})
        )
        return dataset_path_mapping


@pytest.fixture(scope="session")
def data_paths(
    download_all_test_data: dict[str, Path], download_old_test_data: dict[str, Path]
) -> dict[str, Path]:
    return {**download_all_test_data, **download_old_test_data}

####################################################################################################
# these ones are all OLD
# hopefully we can replace them with the new ones
@pytest.fixture(scope="session")
def curr_path() -> str:
    return pathlib.Path(__file__).parent.parent.resolve().as_posix()


@pytest.fixture(scope="session")
def quebec_paths(
    data_paths: dict[str, Path],  data_dir: pathlib.Path
) -> Tuple[str, str, str, str]:
    quebec_path = data_paths["Head-Neck-PET-CT"]
    assert quebec_path.exists(), "Dataset not found"

    output_path = pathlib.Path(data_dir.parent, "tests", "temp")
    quebec_path_str = quebec_path.as_posix()
    crawl_path = data_dir / ".imgtools" / "imgtools_Head-Neck-PET-CT.csv"
    edge_path = data_dir / ".imgtools" / "imgtools_Head-Neck-PET-CT_edges.csv"

    return quebec_path_str, output_path.as_posix(), crawl_path.as_posix(), edge_path.as_posix()


@pytest.fixture(scope="session")
def modalities_path(
    data_paths: dict[str, Path], 
) -> dict[str, str]:
    quebec_patientd_path = data_paths["Head-Neck-PET-CT"] / "HN-CHUS-052"

    assert quebec_patientd_path.exists(), "Dataset not found"

    path = {}
    path["CT"] = pathlib.Path(
        quebec_patientd_path, "08-27-1885-CA ORL FDG TEP POS TX-94629/3.000000-Merged-06362"
    ).as_posix()
    path["RTSTRUCT"] = pathlib.Path(
        quebec_patientd_path,
        "08-27-1885-OrophCB.0OrophCBTRTID derived StudyInstanceUID.-94629/Pinnacle POI-41418",
    ).as_posix()
    path["RTDOSE"] = pathlib.Path(
        quebec_patientd_path,
        "08-27-1885-OrophCB.0OrophCBTRTID derived StudyInstanceUID.-94629/11376",
    ).as_posix()
    path["PT"] = pathlib.Path(
        quebec_patientd_path,
        "08-27-1885-CA ORL FDG TEP POS TX-94629/532790.000000-LOR-RAMLA-44600",
    ).as_posix()
    return path



"""
File Handling Utility.

This module provides functionality to handle files with different actions
such as moving, copying, creating symbolic links, and creating hard links.

Classes
-------
FileAction(Enum)
    Enum for file actions including MOVE, COPY, SYMLINK, and HARDLINK.

Functions
---------
handle_file(source_path: Path, resolved_path: Path, action: FileAction, overwrite: bool) -> None
    Perform specified file operations (move, copy, symlink, hardlink) on a file.

Notes
-----
Symlinks vs. Hardlinks:
- **Symlinks (Symbolic Links)**:
    A symbolic link is a shortcut or reference to another file. It creates a new
    file that points to the target file but does not duplicate the file's data.
    If the target file is moved or deleted, the symlink becomes invalid.
    Example: `ln -s target symlink`

- **Hardlinks**:
    A hard link is an additional reference to the same data on the disk. Both the
    original file and the hard link share the same inode, meaning they are
    indistinguishable. Deleting the original file does not affect the hard link,
    and vice versa.
    Example: `ln target hardlink`

Examples
--------
Move a file:
    >>> from pathlib import (
    ...     Path,
    ... )
    >>> handle_file(
    ...     Path("source.txt"),
    ...     Path("destination.txt"),
    ...     action=FileAction.MOVE,
    ... )

Create a symlink:
    >>> handle_file(
    ...     Path("source.txt"),
    ...     Path("symlink.txt"),
    ...     action=FileAction.SYMLINK,
    ... )

Copy a file:
    >>> handle_file(
    ...     Path("source.txt"),
    ...     Path("copy.txt"),
    ...     action=FileAction.COPY,
    ...     overwrite=True,
    ... )
"""

import shutil
from enum import Enum
from pathlib import Path
from typing import List, Type


class FileAction(Enum):
    MOVE = "move"
    COPY = "copy"
    SYMLINK = "symlink"
    HARDLINK = "hardlink"

    def handle(self, source_path: Path, resolved_path: Path) -> None:
        match self:
            case FileAction.MOVE:
                self.move_file(source_path, resolved_path)
            case FileAction.COPY:
                self.copy_file(source_path, resolved_path)
            case FileAction.SYMLINK:
                self.create_symlink(source_path, resolved_path)
            case FileAction.HARDLINK:
                self.create_hardlink(source_path, resolved_path)

    def move_file(self, source_path: Path, resolved_path: Path) -> None:
        source_path.rename(resolved_path)

    def copy_file(self, source_path: Path, resolved_path: Path) -> None:
        shutil.copy2(
            source_path, resolved_path
        )  # shutil.copy2 preserves metadata

    def create_symlink(self, source_path: Path, resolved_path: Path) -> None:
        try:
            real_source = source_path.resolve(strict=True)
        except (FileNotFoundError, RuntimeError, OSError) as e:
            errmsg = (
                f"Invalid source path {source_path}: possible symlink loop"
            )
            raise RuntimeError(errmsg) from e
        resolved_path.symlink_to(real_source, target_is_directory=False)

    def create_hardlink(self, source_path: Path, resolved_path: Path) -> None:
        try:
            resolved_path.hardlink_to(source_path)
        except OSError as e:
            errmsg = (
                f"Failed to create hard link: {source_path} to {resolved_path}"
            )
            raise RuntimeError(errmsg) from e

    @classmethod
    def validate(cls: Type["FileAction"], action: str) -> "FileAction":
        if not isinstance(action, cls):
            try:
                return cls(action)
            except ValueError as e:
                valid_actions = ", ".join([f"`{a.value}`" for a in cls])
                msg = f"Invalid action: {action}. Must be one of: {valid_actions}"
                raise ValueError(msg) from e
        return action

    @staticmethod
    def choices() -> List[str]:
        """Return a list of valid file actions."""
        return [action.value for action in FileAction]


def handle_file(
    source_path: Path,
    resolved_path: Path,
    action: FileAction | str,
    overwrite: bool = False,
) -> None:
    if not isinstance(action, FileAction):
        action = FileAction.validate(action)

    # Check if the source exists
    if not source_path.exists():
        msg = f"Source does not exist: {source_path}"
        raise FileNotFoundError(msg)

    try:
        if not overwrite and resolved_path.exists():
            msg = f"Destination already exists: {resolved_path}"
            raise FileExistsError(msg)
        # Ensure the parent directory exists and has write permission
        resolved_path.parent.mkdir(parents=True, exist_ok=True)
    except PermissionError as e:
        errmsg = f"Failed to create parent directory or no write permission: {resolved_path.parent}"
        raise PermissionError(errmsg) from e

    action.handle(source_path, resolved_path)



from typing import Tuple, Union

import pytest

from imgtools import (
    Coordinate3D,
    Size3D,
    Spacing3D,
    Vector3D,
)


class TestVector3D:
    """Test suite for Vector3D class."""

    @pytest.mark.parametrize(
        "args, expected",
        [
            # Integer values
            ((1, 2, 3), (1, 2, 3)),
            # Tuple input
            (((4, 5, 6),), (4, 5, 6)),
        ],
    )
    def test_initialization(
        self,
        args: Tuple[int, ...],
        expected: Tuple[int, int, int],
    ) -> None:
        """Test Vector3D initialization with different input types."""
        v = Vector3D(*args)
        assert (v.x, v.y, v.z) == expected
        assert (
            isinstance(v.x, int)
            and isinstance(v.y, int)
            and isinstance(v.z, int)
        )

    @pytest.mark.parametrize(
        "invalid_args",
        [
            (1, 2),  # Too few values
            (1, 2, 3, 4),  # Too many values
        ],
    )
    def test_invalid_initialization(
        self, invalid_args: Tuple[Union[int, str], ...]
    ) -> None:
        """Ensure Vector3D raises errors for invalid inputs."""
        with pytest.raises(ValueError):
            Vector3D(*invalid_args)  # type: ignore

    def test_indexing(self) -> None:
        """Test indexing access via int and str."""
        v = Vector3D(1, 2, 3)
        assert v[0] == 1
        assert v[1] == 2
        assert v[2] == 3
        assert v["x"] == 1
        assert v["y"] == 2
        assert v["z"] == 3
        with pytest.raises(IndexError):
            _ = v[3]
        with pytest.raises(IndexError):
            _ = v["invalid"]

    def test_iteration(self) -> None:
        """Ensure Vector3D is iterable."""
        v = Vector3D(1, 2, 3)
        assert list(v) == [1, 2, 3]

        for i, val in enumerate(v, start=1):
            assert val == i

    def test_repr(self) -> None:
        """Ensure Vector3D has a readable representation."""
        v = Vector3D(1, 2, 3)
        assert repr(v) == "Vector3D(x=1, y=2, z=3)"


class TestSize3D:
    """Test suite for Size3D class."""

    @pytest.mark.parametrize(
        "args, expected",
        [
            # Integer values
            ((1, 2, 3), (1, 2, 3)),
            # Tuple input
            (((4, 5, 6),), (4, 5, 6)),
        ],
    )
    def test_initialization(
        self,
        args: Tuple[int, ...],
        expected: Tuple[int, int, int],
    ) -> None:
        """Test Size3D initialization with different input types."""
        s = Size3D(*args)
        assert (s.width, s.height, s.depth) == expected

    def test_invalid_initialization(self) -> None:
        """Ensure Size3D raises errors for invalid inputs."""
        with pytest.raises(ValueError):
            Size3D(1, 2)

    def test_volume(self) -> None:
        """Ensure Size3D correctly calculates volume."""
        s = Size3D(3, 4, 5)
        assert s.volume == 60

    def test_iteration(self) -> None:
        """Ensure Size3D is iterable."""
        s = Size3D(1, 2, 3)
        assert list(s) == [1, 2, 3]

    def test_repr(self) -> None:
        """Ensure Size3D has a readable representation."""
        s = Size3D(1, 2, 3)
        assert repr(s) == "Size3D(w=1, h=2, d=3)"


class TestCoordinate3D:
    """Test suite for Coordinate3D class."""

    @pytest.mark.parametrize(
        "args, expected",
        [
            # Integer values
            ((1, 2, 3), (1, 2, 3)),
            # Tuple input
            (((4, 5, 6),), (4, 5, 6)),
        ],
    )
    def test_initialization(
        self,
        args: Tuple[int, ...],
        expected: Tuple[int, int, int],
    ) -> None:
        """Test Coordinate3D initialization."""
        c = Coordinate3D(*args)
        assert (c.x, c.y, c.z) == expected

    def test_addition(self) -> None:
        """Test Coordinate3D addition with Size3D."""
        c = Coordinate3D(10, 20, 30)
        s = Size3D(1, 2, 3)
        assert (c + s) == Coordinate3D(11, 22, 33)

        c2 = Coordinate3D(10, 20, 30)
        new = c + c2
        assert new == Coordinate3D(20, 40, 60)

        # we also let users add tuples, just for the sake of it
        tuple_c = (10, 20, 30)
        new_with_tuple = c + tuple_c
        assert new_with_tuple == Coordinate3D(20, 40, 60)

    def test_subtraction(self) -> None:
        """Test Coordinate3D subtraction with Size3D."""
        c = Coordinate3D(10, 20, 30)
        s = Size3D(1, 2, 3)
        assert (c - s) == Coordinate3D(9, 18, 27)

        c2 = Coordinate3D(10, 20, 30)
        new = c - c2
        assert new == Coordinate3D(0, 0, 0)

        # we also let users subtract tuples, just for the sake of it
        tuple_c = (10, 20, 30)
        new_with_tuple = c - tuple_c
        assert new_with_tuple == Coordinate3D(0, 0, 0)

    def test_invalid_operations(self) -> None:
        """Ensure Coordinate3D raises errors for unsupported operations."""
        c = Coordinate3D(10, 20, 30)
        with pytest.raises(TypeError):
            _ = c + "invalid"  # type: ignore
        with pytest.raises(TypeError):
            _ = c - (1, 2)  # type: ignore

    def test_iteration(self) -> None:
        """Ensure Coordinate3D is iterable."""
        c = Coordinate3D(1, 2, 3)
        assert list(c) == [1, 2, 3]


class TestSpacing3D:
    """Test suite for Spacing3D class."""

    @pytest.mark.parametrize(
        "args, expected",
        [
            # Integer values
            ((1, 2, 3), (1, 2, 3)),
            # Tuple input
            (((4, 5, 6),), (4, 5, 6)),
        ],
    )
    def test_initialization(
        self,
        args: Tuple[int, ...],
        expected: Tuple[int, int, int],
    ) -> None:
        """Test Spacing3D initialization."""
        sp = Spacing3D(*args)
        assert (sp.x, sp.y, sp.z) == expected



from pathlib import Path

import click

from imgtools.cli import set_log_verbosity
from imgtools.dicom import find_dicoms
from imgtools.dicom.index import (
    DatabaseHandler,
    DICOMDatabaseInterface,
    DICOMIndexer,
)
from imgtools.logging import logger

DEFAULT_DB_DIR = Path(".imgtools")
DEFAULT_DB_NAME = "imgtools.db"


@click.command()
@set_log_verbosity()
@click.option(
    "--directory",
    "-d",
    type=click.Path(
        exists=True,
        file_okay=False,
        dir_okay=True,
        readable=True,
        resolve_path=True,
        path_type=Path,
    ),
    help="Directory to search for DICOM files",
    required=True,
)
@click.option(
    "--update-db",
    "-u",
    is_flag=True,
    help="Force update the database",
    default=False,
)
@click.option(
    "--limit",
    "-l",
    type=int,
    default=0,
    help="Number of files to index, 0 for all",
)
def index(
    directory: Path,
    update_db: bool = False,
    verbose: int = 0,
    quiet: bool = False,
    limit: int = 0,
) -> None:
    """Index DICOM files in a directory.

    For now, creates a database in the current directory at .imgtools/imgtools.db.

    """
    extension = "dcm"
    check_header = False
    logger.info("Searching for DICOM files.", args=locals())

    dicom_files = find_dicoms(
        directory=directory,
        check_header=check_header,
        recursive=True,
        extension=extension,
    )
    if not dicom_files:
        warningmsg = f"No DICOM files found in {directory}."
        logger.warning(
            warningmsg,
            directory=directory,
            check_header=check_header,
            recursive=True,
            extension=extension,
        )
        return

    logger.info("DICOM find successful.", count=len(dicom_files))

    db_path = DEFAULT_DB_DIR / DEFAULT_DB_NAME

    db_path.parent.mkdir(parents=True, exist_ok=True)

    db_handler = DatabaseHandler(db_path=db_path, force_delete=update_db)

    indexer = DICOMIndexer(db_handler=db_handler)

    if limit:
        dicom_files = dicom_files[:limit]

    logger.debug("Building index.", count=len(dicom_files), limit=limit)
    # Build index
    indexer.build_index_from_files(dicom_files)

    _ = DICOMDatabaseInterface(db_handler=db_handler)

    logger.info("Indexing complete.")

    click.echo(f"Indexed {len(dicom_files)} files to {db_path}")


if __name__ == "__main__":
    index()



from typing import Dict, TypeVar

import SimpleITK as sitk

T = TypeVar("T")


class Scan:
    def __init__(self, image: sitk.Image, metadata: Dict[str, T]) -> None:
        self.image = image
        self.metadata = metadata



import csv
import json
import os
import pathlib
import pickle
import shutil
from datetime import datetime, timezone

import h5py # type: ignore
import nrrd # type: ignore
import numpy as np
import SimpleITK as sitk
from skimage.measure import regionprops

from imgtools.utils import image_to_array

# ruff: noqa
class BaseWriter:
    def __init__(self, root_directory, filename_format, create_dirs=True):
        self.root_directory = root_directory
        self.filename_format = filename_format
        self.create_dirs = create_dirs
        if create_dirs and not os.path.exists(self.root_directory):
            os.makedirs(self.root_directory)

    def put(self, *args, **kwargs):
        raise NotImplementedError

    def _get_path_from_subject_id(self, subject_id, **kwargs):
        now = datetime.now(timezone.utc)
        date = now.strftime("%Y-%m-%d")
        time = now.strftime("%H%M%S")
        date_time = date + "_" + time
        out_filename = self.filename_format.format(
            subject_id=subject_id, date=date, time=time, date_time=date_time, **kwargs
        )
        out_path = pathlib.Path(self.root_directory, out_filename).as_posix()
        out_dir = os.path.dirname(out_path)
        if self.create_dirs and not os.path.exists(out_dir):
            os.makedirs(
                out_dir, exist_ok=True
            )  # create subdirectories if specified in filename_format

        return out_path


class BaseSubjectWriter(BaseWriter):
    def __init__(
        self,
        root_directory,
        filename_format="{subject_id}.nii.gz",
        create_dirs=True,
        compress=True,
    ):
        super().__init__(root_directory, filename_format, create_dirs)
        self.root_directory = root_directory
        self.filename_format = filename_format
        self.create_dirs = create_dirs
        self.compress = compress
        if os.path.exists(self.root_directory):
            # delete the folder called {subject_id} that was made in the original BaseWriter / the one named {label_or_image}
            if os.path.basename(os.path.dirname(self.root_directory)) == "{subject_id}":
                shutil.rmtree(os.path.dirname(self.root_directory))
            elif "{label_or_image}{train_or_test}" in os.path.basename(self.root_directory):
                shutil.rmtree(self.root_directory)

    def put(
        self,
        subject_id,
        image,
        is_mask=False,
        nnunet_info=None,
        label_or_image: str = "images",
        mask_label: str = "",
        train_or_test: str = "Tr",
        **kwargs,
    ):
        if is_mask:
            # remove illegal characters for Windows/Unix
            badboys = r'<>:"/\|?*'
            for char in badboys:
                mask_label = mask_label.replace(char, "")

            # filename_format eh
            self.filename_format = (
                mask_label + ".nii.gz"
            )  # save the mask labels as their rtstruct names

        if nnunet_info:
            if label_or_image == "labels":
                filename = f"{subject_id}.nii.gz"  # naming convention for labels
            else:
                filename = self.filename_format.format(
                    subject_id=subject_id,
                    modality_index=nnunet_info["modalities"][nnunet_info["current_modality"]],
                )  # naming convention for images
            out_path = self._get_path_from_subject_id(
                filename, label_or_image=label_or_image, train_or_test=train_or_test
            )
        else:
            out_path = self._get_path_from_subject_id(self.filename_format, subject_id=subject_id)
        sitk.WriteImage(image, out_path, self.compress)

    def _get_path_from_subject_id(self, filename, **kwargs):
        root_directory = self.root_directory.format(
            **kwargs
        )  # replace the {} with the kwargs passed in from .put() (above)
        out_path = pathlib.Path(root_directory, filename).as_posix()
        out_dir = os.path.dirname(out_path)
        if self.create_dirs and not os.path.exists(out_dir):
            os.makedirs(
                out_dir, exist_ok=True
            )  # create subdirectories if specified in filename_format
        return out_path


class ImageFileWriter(BaseWriter):
    def __init__(
        self,
        root_directory,
        filename_format="{subject_id}.nii.gz",
        create_dirs=True,
        compress=True,
    ):
        super().__init__(root_directory, filename_format, create_dirs)
        self.compress = compress

    def put(self, subject_id, image, **kwargs):
        out_path = self._get_path_from_subject_id(subject_id, **kwargs)
        sitk.WriteImage(image, out_path, self.compress)


class SegNrrdWriter(BaseWriter):
    def __init__(
        self,
        root_directory,
        filename_format="{subject_id}.seg.nrrd",
        create_dirs=True,
        compress=True,
    ):
        super().__init__(root_directory, filename_format, create_dirs)
        if compress:
            self.compression_level = 9
        else:
            self.compression_level = 1

    def put(self, subject_id, mask, **kwargs):
        out_path = self._get_path_from_subject_id(subject_id, **kwargs)
        labels = [k for k in mask.roi_names]
        print(labels)

        origin = mask.GetOrigin()
        spacing = mask.GetSpacing()
        # direction = mask.GetDirection()

        space = "left-posterior-superior"  # everything is ITK read/write

        # fix reverted somewhere.... :''''(
        space_directions = [
            [spacing[0], 0.0, 0.0],
            [0.0, spacing[1], 0.0],
            [0.0, 0.0, spacing[2]],
        ]
        kinds = ["domain", "domain", "domain"]
        dims = 3

        # permute axes to original orientations
        if len(labels) > 1:
            arr = np.transpose(sitk.GetArrayFromImage(mask), [-1, -2, -3, -4])

            # add extra dimension to metadata
            space_directions.insert(0, [float("nan"), float("nan"), float("nan")])
            kinds.insert(0, "vector")
            dims += 1
        else:
            arr = np.transpose(sitk.GetArrayFromImage(mask), [-1, -2, -3])

        # ensure proper conversion to array
        assert mask.GetSize() == arr.shape[-3:]

        segment_info = {}
        for n, i in enumerate(labels):
            try:
                if len(labels) > 1:
                    props = regionprops(arr[n])[0]
                else:
                    props = regionprops(arr)[0]
                bbox = props["bbox"]
                bbox_segment = [bbox[0], bbox[3], bbox[1], bbox[4], bbox[2], bbox[5]]
            except IndexError:  # mask is empty
                assert arr[n].sum() == 0, "Mask not empty but 'skimage.measure.regionprops' failed."
                bbox_segment = [0, 0, 0, 0, 0, 0]

            segment_info[f"Segment{n}_Color"] = list(np.random.random(3))
            segment_info[f"Segment{n}_ColorAutoGenerated"] = "1"
            segment_info[f"Segment{n}_Extent"] = bbox_segment
            segment_info[f"Segment{n}_ID"] = str(n)
            segment_info[f"Segment{n}_Name"] = i
            segment_info[f"Segment{n}_NameautoGenerated"] = "0"

        header = {
            "dimension": dims,
            "space": space,
            "sizes": mask.GetSize(),
            "space directions": space_directions,
            "kinds": kinds,
            "endian": "little",
            "space origin": origin,
            "roi_names": labels,
            **segment_info,
        }

        nrrd.write(
            out_path,
            arr,
            header=header,
            compression_level=self.compression_level,
            **kwargs,
        )


class NumpyWriter(BaseWriter):
    def __init__(self, root_directory, filename_format="{subject_id}.npy", create_dirs=True):
        super().__init__(root_directory, filename_format, create_dirs)
        self.root_directory = root_directory
        self.filename_format = filename_format

    def put(self, subject_id, image, **kwargs):
        out_path = self._get_path_from_subject_id(subject_id, **kwargs)
        if isinstance(image, sitk.Image):
            array, *_ = image_to_array(image)  # TODO (Michal) optionally save the image geometry
        np.save(out_path, array)


class HDF5Writer(BaseWriter):
    def __init__(
        self,
        root_directory,
        filename_format="{subject_id}.h5",
        create_dirs=True,
        save_geometry=True,
    ):
        super().__init__(root_directory, filename_format, create_dirs)
        self.save_geometry = save_geometry

    def put(self, subject_id, images, metadata=None, **kwargs):
        out_path = self._get_path_from_subject_id(subject_id, **kwargs)
        with h5py.File(out_path, "w") as f:
            if not isinstance(images, dict):
                images = {"image": images}
            for k, v in images.items():
                array, origin, direction, spacing = image_to_array(v)
                dataset = f.create_dataset(k, data=array)
                dataset.attrs.create("subject_id", subject_id)
                if self.save_geometry:
                    dataset.attrs.create("origin", data=origin)
                    dataset.attrs.create("direction", data=direction)
                    dataset.attrs.create("spacing", data=spacing)
            if metadata:
                for k, attrs in metadata.items():
                    for name, v in attrs:
                        f[subject_id].attrs.create(name, data=v)


class MetadataWriter(BaseWriter):
    def __init__(
        self,
        root_directory,
        filename_format="{subject_id}.json",
        create_dirs=True,
        remove_existing=True,
    ):
        super().__init__(root_directory, filename_format, create_dirs)
        self.file_format = os.path.splitext(filename_format)[1].lstrip(".")
        self.remove_existing = remove_existing
        if self.file_format not in ["json", "csv", "pkl"]:
            raise ValueError(
                f"File format {self.file_format} not supported. Supported formats: JSON (.json), CSV (.csv), Pickle (.pkl)."
            )

        if self.file_format == "csv" and self.remove_existing:
            out_path = pathlib.Path(self.root_directory, self.filename_format).as_posix()
            if os.path.exists(out_path):
                os.remove(out_path)  # remove existing CSV instead of appending

    def _put_json(self, out_path, **kwargs):
        with open(out_path, "w") as f:
            json.dump(kwargs, f)

    def _put_csv(self, out_path, **kwargs):
        with open(out_path, "a+") as f:
            writer = csv.DictWriter(f, fieldnames=kwargs.keys())
            pos = f.tell()
            f.seek(0)
            sample = "\n".join([f.readline() for _ in range(2)])
            if sample == "\n" or not csv.Sniffer().has_header(sample):
                writer.writeheader()
            f.seek(pos)
            writer.writerow(kwargs)

    def _put_pickle(self, out_path, **kwargs):
        with open(out_path, "wb") as f:
            pickle.dump(kwargs, f)

    def put(self, subject_id, **kwargs):
        out_path = self._get_path_from_subject_id(subject_id)

        if "subject_id" not in kwargs:
            kwargs["subject_id"] = subject_id

        if self.file_format == "json":
            self._put_json(out_path, **kwargs)
        elif self.file_format == "csv":
            self._put_csv(out_path, **kwargs)
        elif self.file_format == "pkl":
            self._put_pickle(out_path, **kwargs)



from .examples import (
    data_images as example_data,
    data_paths as example_data_paths,
)

__all__ = [
    "example_data",
    "example_data_paths",
]


def is_testdata_available() -> bool:
    try:
        from github import Github  # type: ignore # noqa
        import aiohttp  # noqa

        return True
    except ImportError:
        return False


if is_testdata_available():
    from .github_helper import (
        GitHubRelease,
        GitHubReleaseAsset,
        MedImageTestData,
    )

    __all__ += [
        "GitHubRelease",
        "GitHubReleaseAsset",
        "MedImageTestData",
    ]



from .args import parser
from .autopipeutils import save_data
from .dicomutils import all_modalities_metadata, get_modality_metadata
from .imageutils import (
    Array3D,
    ImageArrayMetadata,
    array_to_image,
    idxs_to_physical_points,
    image_to_array,
    physical_points_to_idxs,
)
from .nnunet import (
    generate_dataset_json,
    get_identifiers_from_splitted_files,
    markdown_report_images,
    save_json,
    subfiles,
)

__all__ = [
    "array_to_image",
    "image_to_array",
    "physical_points_to_idxs",
    "idxs_to_physical_points",
    "Array3D",
    "ImageArrayMetadata",
    "get_modality_metadata",
    "all_modalities_metadata",
    "parser",
    "markdown_report_images",
    "save_json",
    "get_identifiers_from_splitted_files",
    "subfiles",
    "generate_dataset_json",
    "save_data",
]



from typing import Dict

import numpy as np


class SparseMask:
    def __init__(
        self, mask_array: np.ndarray, roi_name_dict: Dict[str, int]
    ) -> None:
        self.mask_array = mask_array
        self.roi_name_dict = roi_name_dict



import pathlib
from typing import List

import click


def is_testdata_available() -> bool:
    try:
        from github import Github  # type: ignore # noqa

        return True
    except ImportError:
        return False


if is_testdata_available():
    from imgtools.datasets import MedImageTestData


@click.command(
    no_args_is_help=True,
)
@click.argument(
    "dest",
    type=click.Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
        writable=True,
        path_type=pathlib.Path,
        resolve_path=True,
    ),
)
@click.option(
    "-a",
    "--assets",
    multiple=True,
    help="Specific assets to download. If not provided, all assets will be downloaded.",
)
@click.option("--no", multiple=True, help="Assets to exclude from download.")
@click.help_option(
    "-h",
    "--help",
)
def testdata(dest: pathlib.Path, assets: List[str], no: List[str]) -> None:
    """Download test data from the latest GitHub release.

    DEST is the directory where the test data will be saved.

    assets can be one of the following:

    \b
    - 4D-Lung.tar.gz
    - CC-Tumor-Heterogeneity.tar.gz
    - NSCLC-Radiomics.tar.gz
    - NSCLC_Radiogenomics.tar.gz
    - QIN-PROSTATE-Repeatability.tar.gz
    - Soft-tissue-Sarcoma.tar.gz
    - Vestibular-Schwannoma-SEG.tar.gz

    """
    manager = MedImageTestData()
    selected_assets = None

    if assets:
        selected_assets = [
            asset for asset in manager.datasets if asset.name in assets
        ]
        if not selected_assets:
            click.echo(f"No matching assets found for: {', '.join(assets)}")
            return

    downloaded_files = manager.download(
        dest, assets=selected_assets, exclude=no
    )
    # click.echo(
    #     f"Downloaded files: {', '.join(str(file) for file in downloaded_files)}"
    # )
    click.echo("Downloaded files:")
    for file in downloaded_files:
        click.echo(f"\t{file}")


if __name__ == "__main__":
    testdata()



"""
This code is for testing functioning of different modalities
"""

import os
import pathlib

import pydicom
import pytest
import SimpleITK as sitk

from imgtools.io import read_dicom_auto
from imgtools.ops import StructureSetToSegmentation


@pytest.mark.parametrize('modalities', ['CT', 'RTSTRUCT', 'RTDOSE', 'PT'])
def test_modalities(
    modalities, modalities_path
) -> None:  # modalities_path is a fixture defined in conftest.py
    path = modalities_path
    img = read_dicom_auto(path['CT']).image
    if modalities != 'RTSTRUCT':
        # Checks for dimensions
        dcm = pydicom.dcmread(
            pathlib.Path(path[modalities], os.listdir(path[modalities])[0]).as_posix()
        ).pixel_array
        instances = len(os.listdir(path[modalities]))
        dicom = read_dicom_auto(path[modalities])
        if modalities == 'CT':
            dicom = dicom.image
        if instances > 1:  # For comparing CT and PT modalities
            assert dcm.shape == (dicom.GetHeight(), dicom.GetWidth())
            assert instances == dicom.GetDepth()
        else:  # For comparing RTDOSE modalties
            assert dcm.shape == (dicom.GetDepth(), dicom.GetHeight(), dicom.GetWidth())
        if modalities == 'PT':
            dicom = dicom.resample_pet(img)
            assert dicom.GetSize() == img.GetSize()
        if modalities == 'RTDOSE':
            dicom = dicom.resample_dose(img)
            assert dicom.GetSize() == img.GetSize()
    else:
        struc = read_dicom_auto(path[modalities])
        make_binary_mask = StructureSetToSegmentation(
            roi_names=['GTV.?', 'LARYNX'], continuous=False
        )
        mask = make_binary_mask(struc, img, {'background': 0}, False)
        A = sitk.GetArrayFromImage(mask)
        assert len(A.shape) == 4
        assert A.shape[0:3] == (img.GetDepth(), img.GetHeight(), img.GetWidth())



from .database_handler import DatabaseHandler  # noqa
from .database import DICOMDatabaseInterface, DICOMIndexer, DICOMInsertMixin

__all__ = [
    "DatabaseHandler",
    "DICOMDatabaseInterface",
    "DICOMIndexer",
    "DICOMInsertMixin",
]



from .database import (
    DatabaseHandler,
    DICOMDatabaseInterface,
    DICOMIndexer,
    DICOMInsertMixin,
)

__all__ = [
    "DatabaseHandler",
    "DICOMDatabaseInterface",
    "DICOMIndexer",
    "DICOMInsertMixin",
]



"""
helper_types.py: Intuitive and reusable types for 3D spatial operations.

This module defines foundational types to represent and manipulate 3D spatial
concepts intuitively, focusing on clarity, usability, and alignment with
medical imaging workflows. These types aim to simplify operations like
spatial transformations, bounding box calculations, and metadata representation.

## Goals:
- **Intuitive Design:** Variable names and methods should be easy to understand
  and reflect their real-world meaning.
- **Reusability:** Types should be generic enough to apply to various spatial
  operations across domains, especially medical imaging.

## Types:
1. **Vector3D**
   - Represents a vector in 3D space with x, y, z components.
   - Includes methods for basic vector arithmetic (addition, subtraction).

2. **Size3D**
   - Represents the dimensions of a 3D object (width, height, depth).
   - Includes methods to calculate volume.

3. **Coordinate3D**
   - Represents a specific coordinate in 3D space.
   - Inherits from `Vector3D` and includes methods for addition and subtraction
     with other `Coordinate3D` or `Size3D` objects.

4. **Spacing3D**
   - Represents the spacing in 3D space.
   - Inherits from `Vector3D`.
"""

from __future__ import annotations

from dataclasses import dataclass

from typing import Iterator
import numpy as np


@dataclass
class Vector3D:
    """
    Represent a vector in 3D space.

    Attributes
    ----------
    x :int
        X-component of the vector.
    y : int
        Y-component of the vector.
    z : int
        Z-component of the vector.

    Methods
    -------
    __add__(other):
        Add another Vector3D to this vector.
    __sub__(other):
        Subtract another Vector3D from this vector.
    __iter__():
        Iterate over the components (x, y, z).
    __getitem__(index):
        Access components via index.
    """

    x: int
    y: int
    z: int

    def __init__(self, *args: int) -> None:
        """Initialize a Vector3D with x, y, z components."""
        match args:
            case [x, y, z]:
                self.x, self.y, self.z = x, y, z
            case [tuple_points] if isinstance(tuple_points, tuple):
                self.x, self.y, self.z = tuple_points
            case _:
                errmsg = (
                    f"{self.__class__.__name__} expects 3 values for x, y, z."
                    f" Got {len(args)} values for {args}."
                )
                raise ValueError(errmsg)

    def __iter__(self) -> Iterator[int]:
        """Allow iteration over the components."""
        return iter((self.x, self.y, self.z))

    def __getitem__(self, idx: int | str) -> int:
        """Access components via index."""
        match idx:
            case int() as idx:
                return (self.x, self.y, self.z)[idx]
            case str() if idx in vars(self):
                return getattr(self, idx)
            case _:
                errmsg = f"Invalid index: {idx}."
                raise IndexError(errmsg)

    def __repr__(self) -> str:
        """Return a string representation of the Vector3D."""
        cls = self.__class__.__name__
        return f"{cls}(x={self.x}, y={self.y}, z={self.z})"


class Coordinate3D(Vector3D):
    """
    Represent a point in 3D space.
    Inherits from Vector3D.

    Can add and subtract other Point3D or Size3D objects.
    """

    def __add__(self, other: Coordinate3D | Size3D | tuple) -> Coordinate3D:
        """Add another Vector3D to this vector."""
        match other:
            case Coordinate3D(x, y, z):
                return Coordinate3D(self.x + x, self.y + y, self.z + z)
            case Size3D(width, height, depth):
                return Coordinate3D(
                    self.x + width, self.y + height, self.z + depth
                )
            case tuple() if len(other) == 3:
                return Coordinate3D(
                    self.x + other[0], self.y + other[1], self.z + other[2]
                )
            case _:
                errmsg = (
                    f"Unsupported type for addition: {type(other)}."
                    " Expected Coordinate3D or Size3D."
                )
                raise TypeError(errmsg)

    def __sub__(self, other: Coordinate3D | Size3D | tuple) -> Coordinate3D:
        """Subtract another Vector3D from this vector."""
        match other:
            case Coordinate3D(x, y, z):
                return Coordinate3D(self.x - x, self.y - y, self.z - z)
            case Size3D(width, height, depth):
                return Coordinate3D(
                    self.x - width, self.y - height, self.z - depth
                )
            case tuple() if len(other) == 3:
                return Coordinate3D(
                    self.x - other[0], self.y - other[1], self.z - other[2]
                )
            case _:
                errmsg = (
                    f"Unsupported type for subtraction: {type(other)}."
                    " Expected Coordinate3D or Size3D."
                )
                raise TypeError(errmsg)


class Spacing3D:
    """
    Represent the spacing in 3D space.
    Inherits from Vector3D.
    """

    x: float
    y: float
    z: float

    def __init__(self, *args: float) -> None:
        """Initialize a Spacing3D with x, y, z components."""
        match args:
            case [x, y, z]:
                self.x, self.y, self.z = x, y, z
            case [tuple_points] if isinstance(tuple_points, tuple):
                self.x, self.y, self.z = tuple_points
            case _:
                errmsg = (
                    f"{self.__class__.__name__} expects 3 values for x, y, z."
                    f" Got {len(args)} values for {args}."
                )
                raise ValueError(errmsg)

    def __repr__(self) -> str:
        """Return a string representation of the Spacing3D."""
        return f"Spacing3D(x={self.x}, y={self.y}, z={self.z})"

    def __iter__(self) -> Iterator[float]:
        """Allow iteration over the components."""
        return iter((self.x, self.y, self.z))


@dataclass
class Size3D:
    """
    Represent the size (width, height, depth) of a 3D object.

    Attributes
    ----------
    width : int
        The width of the 3D object.
    height : int
        The height of the 3D object.
    depth : int
        The depth of the 3D object.

    Methods
    -------
    volume:
        Calculate the volume of the 3D object.
    """

    width: int
    height: int
    depth: int

    def __init__(self, *args: int) -> None:
        match args:
            case [int() as width, int() as height, int() as depth]:
                self.width, self.height, self.depth = width, height, depth
            case [tuple_points] if isinstance(tuple_points, tuple):
                self.width, self.height, self.depth = map(int, tuple_points)
            case _:
                errmsg = (
                    f"{self.__class__.__name__} expects 3 values for width, height, depth."
                    f" Got {len(args)} values for {args}."
                )
                raise ValueError(errmsg)

    @property
    def volume(self) -> int:
        """Calculate the volume of the 3D object."""
        return self.width * self.height * self.depth

    def __iter__(self) -> Iterator[int]:
        """Allow iteration over the dimensions."""
        return iter((self.width, self.height, self.depth))

    def __repr__(self) -> str:
        """Return a string representation of the Size3D."""
        return f"Size3D(w={self.width}, h={self.height}, d={self.depth})"


if __name__ == "__main__":  # pragma: no cover
    # ruff : noqa
    from pathlib import Path
    from rich import print

    ########################################
    # Vector3D
    ########################################
    myint = 10
    print(myint)

    vector1 = Vector3D(1, 2, 3)

    # as tuple input
    vector2 = Vector3D(*(1, 2, 3))
    assert all((attr1 == attr2) for attr1, attr2 in zip(vector1, vector2))

    # iterate over the objects' attributes
    for attr in vector1:
        print(attr)

    print(vector1)

    # access object attributes via [ ]
    print(f"{vector1[0]}")
    print(f"{vector1['y']}")

    ########################################
    # Coordinate3D
    ########################################

    point = Coordinate3D(10, 20, 30)
    point2 = Coordinate3D((15, 25, 35)) # mypy will complain about this, but it will work
    point2 = Coordinate3D(*(15, 25, 35)) # unpack the tuple to avoid mypy error

    print(point)
    print(point2)

    size_tuple = (50, 60, 70)
    size = Size3D(size_tuple) # mypy will complain about this, but it will work
    size = Size3D(*size_tuple) # unpack the tuple
    point_plus_size = point + size_tuple

    print(f"Adding {point=} and {size_tuple=} = {point_plus_size}")
    print(f"Adding {point=} and {size=} = {point + size}")


    # make a sitk image
    import SimpleITK as sitk

    # create a 3D image
    image = sitk.Image(100, 100, 100, sitk.sitkInt16)

    image_size = Size3D(*image.GetSize())
    print(f"Image size: {image_size}")

    # get the spacing
    image_spacing = Spacing3D(*image.GetSpacing())
    print(f"Image spacing: {image_spacing}")

    


from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

# from imgtools.sort import lookup_tag, similar_tags, tag_exists
# from imgtools.sort.exceptions import InvalidPatternError, SorterBaseError
# from imgtools.sort.sorter_base import SorterBase
from imgtools.dicom import find_dicoms, lookup_tag, similar_tags, tag_exists
from imgtools.dicom.utils import find_dicoms

########################################################################
# Test Helpers
########################################################################


@pytest.fixture
def temp_dir_with_files():
    with TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)
        # Create dummy DICOM files
        (temp_path / 'file1.dcm').touch()
        (temp_path / 'file2.dcm').touch()
        # Create a subdirectory with more DICOM files
        sub_dir = temp_path / 'subdir'
        sub_dir.mkdir()
        (sub_dir / 'file3.dcm').touch()
        (sub_dir / 'file4.dcm').touch()
        yield temp_path


class TestFindDicoms:
    def test_non_recursive_no_header_check(self, temp_dir_with_files) -> None:
        result = find_dicoms(temp_dir_with_files, recursive=False, check_header=False)
        assert len(result) == 2
        assert all(file.suffix == '.dcm' for file in result)

    def test_recursive_no_header_check(self, temp_dir_with_files) -> None:
        result = find_dicoms(temp_dir_with_files, recursive=True, check_header=False)
        assert len(result) == 4
        assert all(file.suffix == '.dcm' for file in result)

    def test_non_recursive_with_header_check(self, temp_dir_with_files, mocker) -> None:
        mocker.patch('imgtools.dicom.utils.is_dicom', return_value=True)
        result = find_dicoms(temp_dir_with_files, recursive=False, check_header=True)
        assert len(result) == 2
        assert all(file.suffix == '.dcm' for file in result)

    def test_recursive_with_header_check(self, temp_dir_with_files, mocker) -> None:
        mocker.patch('imgtools.dicom.utils.is_dicom', return_value=True)
        result = find_dicoms(temp_dir_with_files, recursive=True, check_header=True)
        assert len(result) == 4
        assert all(file.suffix == '.dcm' for file in result)

    def test_search_with_specific_extension(self, temp_dir_with_files) -> None:
        (temp_dir_with_files / 'file5.txt').touch()
        result = find_dicoms(
            temp_dir_with_files, recursive=True, check_header=False, extension='dcm'
        )
        assert len(result) == 4
        assert all(file.suffix == '.dcm' for file in result)

    def test_limit_results(self, temp_dir_with_files) -> None:
        result = find_dicoms(temp_dir_with_files, recursive=True, check_header=False, limit=2)
        assert len(result) == 2
        assert all(file.suffix == '.dcm' for file in result)

    def test_search_with_input(self, temp_dir_with_files) -> None:
        (temp_dir_with_files / 'scan1.dcm').touch()
        (temp_dir_with_files / 'scan2.dcm').touch()
        result = find_dicoms(
            temp_dir_with_files, recursive=True, check_header=False, search_input=['scan']
        )
        assert len(result) == 2
        assert all('scan' in file.name for file in result)

    def test_combined_options(self, temp_dir_with_files, mocker) -> None:
        mocker.patch('imgtools.dicom.utils.is_dicom', return_value=True)
        (temp_dir_with_files / 'scan1.dcm').touch()
        result = find_dicoms(
            temp_dir_with_files,
            recursive=True,
            check_header=True,
            extension='dcm',
            limit=3,
            search_input=['scan'],
        )
        assert len(result) == 1
        assert all(file.suffix == '.dcm' for file in result)
        assert all('scan' in file.name for file in result)


def test_similar_tags() -> None:
    """Test the similar_tags function."""
    incorrect_2_correct_mappings = {
        'PatientID': 'Patient',
        'StudyInstanceUID': 'StudyIntance',
        'SeriesInstanceUID': 'SeriesIntance',
        'SOPInstanceUID': 'SOPInstance',
        'Modality': 'Modaliti',
        'SeriesDate': 'SerieDate',
    }
    for correct, incorrect in incorrect_2_correct_mappings.items():
        assert correct in similar_tags(incorrect)


def test_lookup_tag() -> None:
    """Test the lookup_tag function."""
    assert lookup_tag('PatientID') == '1048608'
    assert lookup_tag('PatientID', hex_format=True) == '0x100020'
    assert lookup_tag('invalid') is None


def test_tag_exists() -> None:
    """Test the tag_exists function."""
    assert tag_exists('PatientID')



from __future__ import annotations

from dataclasses import dataclass, field

import SimpleITK as sitk

from imgtools.logging import logger

from .helper_types import Coordinate3D, Size3D


# Exception for when the bounding box is outside the image
class BoundingBoxOutsideImageError(Exception):
    """Exception raised when the bounding box is outside the image."""

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


def calculate_image_boundaries(image: sitk.Image) -> RegionBox:
    """
    Calculate the physical coordinate boundaries of a SimpleITK image.

    Parameters
    ----------
    image: sitk.Image
        The input SimpleITK image.

    Returns
    -------
    RegionBox

    Examples
    --------
    >>> calculate_image_boundaries(image)
    """

    min_coord = Coordinate3D(
        *image.TransformPhysicalPointToIndex(image.GetOrigin())
    )
    size = Size3D(*image.GetSize())

    return RegionBox(min_coord, min_coord + size)


@dataclass
class RegionBox:
    """Represents a box in 3D space.

    Attributes
    ----------
    min : Coordinate3D
        The minimum coordinate of the box.
    max : Coordinate3D
        The maximum coordinate of the box.

    size : Size3D
        The size of the box, calculated from the min and max coordinates.
    """

    min: Coordinate3D
    max: Coordinate3D
    size: Size3D = field(init=False)

    @classmethod
    def from_tuple(
        cls, coordmin: tuple[int, int, int], coordmax: tuple[int, int, int]
    ) -> RegionBox:
        """Creates a RegionBox from a tuple of min and max coordinates."""
        return cls(Coordinate3D(*coordmin), Coordinate3D(*coordmax))

    @classmethod
    def from_mask_centroid(cls, mask: sitk.Image, label: int = 1) -> RegionBox:
        """Creates a RegionBox from the centroid of a mask image.

        Parameters
        ----------
        mask : sitk.Image
            The input mask image.

        Returns
        -------
        RegionBox
            The bounding box coordinates as a RegionBox object.
        """
        mask_uint = sitk.Cast(mask, sitk.sitkUInt8)
        stats = sitk.LabelShapeStatisticsImageFilter()
        stats.Execute(mask_uint)

        centroid = stats.GetCentroid(label)
        centroid_idx = mask.TransformPhysicalPointToIndex(centroid)

        return RegionBox(
            Coordinate3D(*centroid_idx), Coordinate3D(*centroid_idx)
        )

    @classmethod
    def from_mask_bbox(cls, mask: sitk.Image, label: int = 1) -> RegionBox:
        """Creates a RegionBox from the bounding box of a mask image.

        Parameters
        ----------
        mask : sitk.Image
            The input mask image.
        label : int, optional

        Returns
        -------
        RegionBox
            The bounding box coordinates as a RegionBox object.
        """

        mask_uint = sitk.Cast(mask, sitk.sitkUInt8)
        stats = sitk.LabelShapeStatisticsImageFilter()
        stats.Execute(mask_uint)
        xstart, ystart, zstart, xsize, ysize, zsize = stats.GetBoundingBox(
            label
        )

        return RegionBox(
            Coordinate3D(xstart, ystart, zstart),
            Coordinate3D(xstart + xsize, ystart + ysize, zstart + zsize),
        )

    def __post_init__(self) -> None:
        if (
            self.min.x > self.max.x
            or self.min.y > self.max.y
            or self.min.z > self.max.z
        ):
            msg = "The minimum coordinate must be less than the maximum coordinate."
            msg += f" Got: min={self.min}, max={self.max}"
            raise ValueError(msg)

        self.size = Size3D(
            self.max.x - self.min.x,
            self.max.y - self.min.y,
            self.max.z - self.min.z,
        )

    def __repr__(self) -> str:
        """prints out like this:

        RegionBox(
            min=Coordinate3D(x=223, y=229, z=57),
            max=Coordinate3D(x=303, y=299, z=87)
            size=(80, 70, 30)
        )
        """
        return (
            f"{self.__class__.__name__}(\n"
            f"\tmin={self.min},\n"
            f"\tmax={self.max}\n"
            f"\tsize={self.size}\n"
            f")"
        )

    def pad(self, padding: int) -> RegionBox:
        """Expand the bounding box by a specified padding value in all directions.

        Parameters
        ----------
        padding : int
            The padding value to expand the bounding box.

        Returns
        -------
        RegionBox
            The expanded bounding box.
        """
        if padding == 0:
            return self

        padded_min = Coordinate3D(
            self.min.x - padding,
            self.min.y - padding,
            self.min.z - padding,
        )
        padded_max = Coordinate3D(
            self.max.x + padding,
            self.max.y + padding,
            self.max.z + padding,
        )
        return RegionBox(min=padded_min, max=padded_max)

    def expand_to_cube(self, desired_size: int | None = None) -> RegionBox:
        """Convert the bounding box to a cube by making the size equal along all dimensions.

        This is done by finding which dimension is the largest,
        and then pad the other dimensions to make them equal to the desired size.

        Parameters
        ----------
        desired_size : int | None
            The desired size of the cube. If None, the maximum dimension size is used

        Returns
        -------
        RegionBox
            The bounding box converted to a cube.

        Raises
        ------
        ValueError
            If the desired size is smaller than the current maximum dimension size.
        """
        max_size = max(self.size)

        if not desired_size:
            return self.minimum_dimension_size(max_size)

        if desired_size < max_size:
            msg = (
                f"Desired size {desired_size} is smaller than"
                f" the current maximum dimension size {max_size}."
            )
            raise ValueError(msg)

        return self.minimum_dimension_size(desired_size)

    def minimum_dimension_size(self, size: int = 5) -> RegionBox:
        """Ensure that the bounding box has a minimum size along each dimension.

        Parameters
        ----------
        size : int
            The minimum size of the bounding box along each dimension.

        Returns
        -------
        RegionBox
            The bounding box with a minimum size along each dimension.

        Notes
        -----
        Validation is done to ensure that any min coordinates that are negative are set to 0,
        and the difference is added to the maximum coordinates
        """
        extra_x = max(0, size - self.size.width) // 2
        extra_y = max(0, size - self.size.height) // 2
        extra_z = max(0, size - self.size.depth) // 2

        min_coord = self.min - (extra_x, extra_y, extra_z)
        max_coord = self.max + (extra_x, extra_y, extra_z)

        self._adjust_negative_coordinates(min_coord, max_coord)

        return RegionBox(min=min_coord, max=max_coord)

    def _adjust_negative_coordinates(
        self, min_coord: Coordinate3D, max_coord: Coordinate3D
    ) -> None:
        """Adjust the coordinates to ensure that the min values are not negative."""
        # if any of the min values are negative, set them to 0,
        # and add the difference to the max values
        for axis in ["x", "y", "z"]:
            min_value = getattr(min_coord, axis)
            if min_value < 0:
                diff = abs(min_value)
                setattr(min_coord, axis, 0)
                setattr(max_coord, axis, getattr(max_coord, axis) + diff)

    def crop_image(self, image: sitk.Image) -> sitk.Image:
        """Crop an image to the coordinates defined by the box.

        Parameters
        ----------
        image : sitk.Image
            The image to crop.

        Returns
        -------
        sitk.Image
            The cropped image.
        """
        try:
            cropped_image = sitk.RegionOfInterest(image, self.size, self.min)
        except RuntimeError as e:
            msg = (
                f"The box {self} is outside"
                f" the image size {calculate_image_boundaries(image)} "
            )
            # this is probably due to the box being outside the image
            # try to crop the image to the largest possible region
            # we could handle this in the future, for now let it raise the error
            # and make user try again with an adjusted box
            raise BoundingBoxOutsideImageError(msg) from e
        else:
            return cropped_image


if __name__ == "__main__":
    from rich import print  # noqa

    basicbox = RegionBox(Coordinate3D(5, 5, 5), Coordinate3D(10, 10, 10))

    print(basicbox)

    tupleinitbox = RegionBox.from_tuple((5, 5, 5), (10, 10, 10))

    print(tupleinitbox)

    # create a 3D image
    image = sitk.Image(100, 100, 100, sitk.sitkInt16)

    cropped_image = basicbox.crop_image(image)

    print(f"{cropped_image.GetSize()=}")

    non_uniform_box = RegionBox(
        Coordinate3D(5, 5, 5), Coordinate3D(21, 23, 29)
    )

    print(f"{non_uniform_box=}")

    padded_box = non_uniform_box.pad(5)

    print(f"{padded_box=}")

    cube_box = non_uniform_box.expand_to_cube()

    print(f"{cube_box=}")

    min_size_box = non_uniform_box.minimum_dimension_size(30)
    print(f"{min_size_box=}")

    print(f"{non_uniform_box.expand_to_cube().minimum_dimension_size(50)=}")

    try:
        box_outsided_image = RegionBox(
            Coordinate3D(0, 0, 0), Coordinate3D(200, 200, 200)
        )

        cropped_image = box_outsided_image.crop_image(image)

        print(f"{cropped_image.GetSize()=}")
    except BoundingBoxOutsideImageError as e:
        print("ERROR")
        print(e)

    ########################################
    # from masks
    ########################################

    ct_image = sitk.ReadImage(
        "/home/bioinf/bhklab/radiomics/readii-negative-controls/rawdata/HEAD-NECK-RADIOMICS-HN1/images/niftis/SubjectID-100_HN1339/CT_82918_original.nii.gz"
    )

    print(f"{calculate_image_boundaries(ct_image)=}")

    rt_image = sitk.ReadImage(
        "/home/bioinf/bhklab/radiomics/readii-negative-controls/rawdata/HEAD-NECK-RADIOMICS-HN1/images/niftis/SubjectID-100_HN1339/RTSTRUCT_11267_GTV.nii.gz"
    )

    print(f"{RegionBox.from_mask_bbox(rt_image)=}")

    print(f"{RegionBox.from_mask_centroid(rt_image)=}")

    print(f"{RegionBox.from_mask_centroid(rt_image).expand_to_cube(50)=}")

    print(
        f"{RegionBox.from_mask_centroid(rt_image).expand_to_cube(50).minimum_dimension_size(30)=}"
    )

    ## assign
    centroid_region = RegionBox.from_mask_centroid(rt_image).expand_to_cube(
        100
    )
    bbox_region = RegionBox.from_mask_bbox(rt_image).expand_to_cube(100)

    cropped_centroid_image = centroid_region.crop_image(ct_image)
    cropped_bbox_image = bbox_region.crop_image(ct_image)

    for name, img in zip(
        ("cropped_centroid_image", "cropped_bbox_image"),
        (cropped_centroid_image, cropped_bbox_image),
        strict=True,
    ):
        print("*" * 20)
        print(f"{name}")
        print(f"{img.GetSize()=}")
        print(f"{img.GetOrigin()=}")
        print(f"{img.GetSpacing()=}")
        print(f"{img.GetDirection()=}")



from .old_loaders import (
    BaseLoader,
    ImageCSVLoader,
    ImageFileLoader,
    ImageTreeLoader,
    auto_dicom_result,
    read_dicom_auto,
    read_dicom_pet,
    read_dicom_rtdose,
    read_dicom_rtstruct,
    read_dicom_scan,
    read_dicom_seg,
    read_dicom_series,
    read_image,
)

__all__ = [
    "read_image",
    "read_dicom_series",
    "read_dicom_scan",
    "read_dicom_rtstruct",
    "read_dicom_rtdose",
    "read_dicom_pet",
    "read_dicom_seg",
    "read_dicom_auto",
    "auto_dicom_result",
    "BaseLoader",
    "ImageTreeLoader",
    "ImageCSVLoader",
    "ImageFileLoader",
]



from __future__ import annotations

from dataclasses import dataclass
from typing import Iterator, Tuple, TypeAlias

import numpy as np

"""
jermiah (01/30/2025): 
not sure about this direction class, was an initial proof-of-concept
doesnt seem like a dealbreaker to need, havent dealt with any issues
Really only wanted to use it for docs/tutorials to show how images
are oriented + work via directions for sitk.Image

lets not consider it too much unless we need to debug with it
"""

Matrix3D: TypeAlias = Tuple[
    Tuple[float, float, float],
    Tuple[float, float, float],
    Tuple[float, float, float],
]
Matrix3DFlat: TypeAlias = Tuple[
    float, float, float, float, float, float, float, float, float
]

FlattenedMatrix = Matrix3DFlat


@dataclass(frozen=True, eq=True)
class Direction:
    """
    Represent a directional matrix for image orientation.

    Supports 3D (3x3) directional matrices in row-major format.

    Attributes
    ----------
    matrix : Tuple[float, ...]
        A flattened 1D array representing the matrix,
        with length 9 (3x3).
    """

    matrix: Matrix3DFlat

    def __post_init__(self) -> None:
        length = len(self.matrix)
        if length != 9:
            msg = (
                "Direction must be a 3x3 (9 values) matrix."
                f" Got {length} values."
            )
            raise ValueError(msg)

    @classmethod
    def from_matrix(
        cls,
        matrix: Matrix3D,
    ) -> Direction:
        """
        Create a Direction object from a full 3D (3x3) matrix.

        Parameters
        ----------
        matrix : Matrix3D
            A nested tuple representing the 3D matrix.

        Returns
        -------
        Direction
            A Direction instance.
        """
        if (size := len(matrix)) != 3:
            msg = f"Matrix must be 3x3. Got {size=}."
            raise ValueError(msg)
        for row in matrix:
            if len(row) != size:
                raise ValueError("Matrix must be square (3x3).")
        flattened: FlattenedMatrix = tuple(
            value for row in matrix for value in row
        )  # type: ignore
        return cls(matrix=flattened)

    def to_matrix(self) -> list[list[float]]:
        """Convert the flattened row-major array back to a 3D matrix."""
        dim = 3
        return [list(self.matrix[i * dim : (i + 1) * dim]) for i in range(dim)]

    def normalize(self) -> Direction:
        """Return a new Direction with normalized row vectors."""
        matrix = self.to_matrix()
        normalized_matrix = [
            list(np.array(row) / np.linalg.norm(row)) for row in matrix
        ]
        return Direction.from_matrix(
            tuple(tuple(row) for row in normalized_matrix)  # type: ignore
        )

    def is_normalized(self, tol: float = 1e-6) -> bool:
        """Check if the row vectors of the matrix are normalized."""
        matrix = self.to_matrix()
        for row in matrix:
            if not np.isclose(np.linalg.norm(row), 1.0, atol=tol):
                return False
        return True

    def __iter__(self) -> Iterator:
        """Allow the Direction instance to be passed directly as a 1D array."""
        yield from self.matrix

    def __repr__(self) -> str:
        dim = 3
        rows = self.to_matrix()
        formatted_rows = [
            "  [" + ", ".join(f"{value:>7.3f}" for value in row) + "]"
            for row in rows
        ]
        return (
            f"Direction(  {dim}x{dim} matrix\n"
            + "\n".join(formatted_rows)
            + "\n)"
        )

    # Create instances of each class
    # point = Point3D(10.0, 20.0, 30.0)
    # size = Size3D(50.0, 60.0, 70.0)
    # direction = Direction.from_matrix(
    #     (
    #         (0.707, 0.707, 0.0),
    #         (-0.707, 0.707, 0.0),
    #         (0.0, 0.0, 1.0),
    #     )
    # )

    # # Testing Point3D and Size3D operations
    # new_point = point + size

    # # Printing out the details
    # print(f"Point: {point}")
    # print(f"Size: {size}")
    # print(f"New point after adding size: {new_point}")
    # print(f"Direction: {direction}")

    # # Unpacking the values
    # print("Unpacked Point:", tuple(point))  # (10.0, 20.0, 30.0)
    # print("Unpacked Size:", tuple(size))  # (50.0, 60.0, 70.0)

    # example_image = np.random.rand(10, 10, 10)
    # example_sitk_image = sitk.GetImageFromArray(example_image)

    # # Create a direction vector
    # # 3x3 Direction Matrix
    # direction_3d = Direction.from_matrix(
    #     (
    #         (0.707, 0.707, 0.0),
    #         (-0.707, 0.707, 0.0),
    #         (0.0, 0.0, 1.0),
    #     )
    # )

    # print(f"{direction_3d=}")
    # example_sitk_image.SetDirection(direction_3d)

    # # make another direction from a flattened
    # direction_3d_flat = Direction(
    #     matrix=(0.707, 0.707, 0.0, -0.707, 0.707, 0.0, 0.0, 0.0, 1.0)
    # )
    # print(f"{direction_3d_flat=}")
    # print(f"{(direction_3d_flat==direction_3d)=}")

    # print(f"{example_sitk_image=}")



"""
OLD BBOX CODE TO GET INSPIRATION FROM
"""

from __future__ import annotations

from dataclasses import dataclass, field

import SimpleITK as sitk

from imgtools.logging import logger


@dataclass
class Point3D:
    """Represent a point in 3D space."""

    x: int
    y: int
    z: int

    @property
    def as_tuple(self) -> tuple[int, int, int]:
        return self.x, self.y, self.z

    def __add__(self, other: Point3D) -> Point3D:
        return Point3D(
            x=self.x + other.x, y=self.y + other.y, z=self.z + other.z
        )

    def __sub__(self, other: Point3D) -> Point3D:
        return Point3D(
            x=self.x - other.x, y=self.y - other.y, z=self.z - other.z
        )


@dataclass
class Size3D(Point3D):
    """Represent the size of a 3D object using its width, height, and depth."""

    pass


@dataclass
class Coordinate(Point3D):
    """Represent a coordinate in 3D space."""

    pass


@dataclass
class Centroid(Coordinate):
    """Represent the centroid of a region in 3D space.

    A centroid is simply a coordinate in 3D space that represents
    the center of mass of a region in an image. It is represented
    by its x, y, and z coordinates.

    Attributes
    ----------
    x : int
    y : int
    z : int
    """

    pass


class BoundingBoxError(Exception):
    pass


def calculate_image_boundaries(image: sitk.Image) -> BoundingBox:
    """
    Calculate the physical coordinate boundaries of a SimpleITK image.

    Parameters
    ----------
    image : sitk.Image
            The input SimpleITK image.

    Returns
    -------
    BoundingBox

    Examples
    --------
    >>> calculate_image_boundaries(image)
    BoundingBox(min=Coordinate(x=0, y=0, z=0), max=Coordinate(x=512, y=512, z=512))
    """

    size = image.GetSize()
    origin = image.GetOrigin()

    min_coord = Coordinate(
        x=origin[0],
        y=origin[1],
        z=origin[2],
    )
    max_coord = Coordinate(
        x=origin[0] + size[0],
        y=origin[1] + size[1],
        z=origin[2] + size[2],
    )

    return BoundingBox(min=min_coord, max=max_coord)


@dataclass
class BoundingBox:
    """
    Represents a rectangular region in a coordinate space.

    Attributes
    ----------
    min : Coordinate
        The minimum coordinate (bottom-left corner) of the bounding box.
    max : Coordinate
        The maximum coordinate (top-right corner) of the bounding box.
    """

    min: Coordinate
    max: Coordinate
    size: Size3D = field(init=False)

    def __post_init__(self):
        if (
            self.min.x > self.max.x
            or self.min.y > self.max.y
            or self.min.z > self.max.z
        ):
            msg = "The minimum coordinate must be less than the maximum coordinate."
            msg += f" Got: min={self.min.as_tuple}, max={self.max.as_tuple}"
            raise ValueError(msg)

        self.size = Size3D(
            x=self.max.x - self.min.x,
            y=self.max.y - self.min.y,
            z=self.max.z - self.min.z,
        )

    def __repr__(self):
        """
        prints out like this:

        BoundingBox(
            min=Coordinate(x=223, y=229, z=57),
            max=Coordinate(x=303, y=299, z=87)
            size=(80, 70, 30)
        )
        """
        return (
            f"{self.__class__.__name__}(\n"
            f"\tmin={self.min},\n"
            f"\tmax={self.max}\n"
            f"\tsize={self.size}\n"
            f")"
        )

    @classmethod
    def from_centroid(
        cls, mask: sitk.Image, size: Size3D, label: int | None = None
    ) -> BoundingBox:
        """Create a bounding box around the centroid of a mask with a given size.

        Parameters
        ----------
        mask : sitk.Image
            The binary mask image.
        size : Size3D
            The size of the bounding box.
        label : int | None
            The label of the region to find the bounding box for.
            if None, the label is determined automatically. Default is None.

        Returns
        -------
        BoundingBox
            The bounding box coordinates as a BoundingBox object.

        Examples
        --------
        >>> size = Size3D(x=5, y=5, z=5)
        >>> find_bbox_from_centroid(mask, size, label=1)
        BoundingBox(min=Coordinate(x=7, y=7, z=7), max=Coordinate(x=12, y=12, z=12))
        """
        if isinstance(size, Size3D):
            pass
        elif isinstance(size, tuple):
            size = Size3D(*size)
        else:
            msg = "Size must be a Size3D object or a tuple of (x, y, z) integers."
            raise ValueError(msg)

        # label = label or getROIVoxelLabel(mask)

        mask_uint = sitk.Cast(mask, sitk.sitkUInt8)
        stats = sitk.LabelShapeStatisticsImageFilter()
        stats.Execute(mask_uint)

        if not stats.HasLabel(label):
            raise ValueError(
                f"The mask does not contain any labeled regions with label {label}."
            )

        centroid_coords = stats.GetCentroid(label)
        centroid_idx = mask.TransformPhysicalPointToIndex(centroid_coords)
        centroid = Centroid(
            x=centroid_idx[0], y=centroid_idx[1], z=centroid_idx[2]
        )

        min_coord = Coordinate(
            x=centroid.x - size.x // 2,
            y=centroid.y - size.y // 2,
            z=centroid.z - size.z // 2,
        )
        max_coord = Coordinate(
            x=centroid.x + size.x // 2,
            y=centroid.y + size.y // 2,
            z=centroid.z + size.z // 2,
        )
        return cls(min_coord, max_coord)

    @classmethod
    def from_mask(
        cls, mask: sitk.Image, min_dim_size: int = 4, pad: int = 0
    ) -> BoundingBox:
        """Find the bounding box of a given mask image.

        Parameters
        ----------
        mask : sitk.Image
            The input mask image.
        min_dim_size : int
            Minimum size of bounding box along each dimension. Default is 4.

        Returns
        -------
        BoundingBox
            The bounding box coordinates as a BoundingBox object.

        Examples
        --------
        >>> find_bbox(mask, min_dim_size=4)
        BoundingBox(min=Coordinate(x=0, y=0, z=0), max=Coordinate(x=4, y=4, z=4))
        """

        mask_uint = sitk.Cast(mask, sitk.sitkUInt8)
        stats = sitk.LabelShapeStatisticsImageFilter()
        stats.Execute(mask_uint)
        xstart, ystart, zstart, xsize, ysize, zsize = stats.GetBoundingBox(1)

        # Ensure minimum size of 4 pixels along each dimension
        xsize = max(xsize, min_dim_size)
        ysize = max(ysize, min_dim_size)
        zsize = max(zsize, min_dim_size)

        min_coord = Coordinate(x=xstart, y=ystart, z=zstart)
        max_coord = Coordinate(
            x=xstart + xsize, y=ystart + ysize, z=zstart + zsize
        )

        if pad == 0:
            return cls(min_coord, max_coord)

        return cls(min_coord, max_coord).pad(padding=pad)

    def crop_image(self, image: sitk.Image) -> sitk.Image:
        """Crop the input image to the bounding box.

        Parameters
        ----------
        image : sitk.Image
            The input image to crop.

        Returns
        -------
        sitk.Image
            The cropped image.
        """
        try:
            cropped_image = sitk.RegionOfInterest(
                image,
                self.size.as_tuple,
                self.min.as_tuple,
            )
        except RuntimeError:
            # this is probably due to the bounding box being outside the image
            # try to crop the image to the largest possible region
            msg = f"The bounding box {self} is outside the image size {calculate_image_boundaries(image)} "
            msg += "Chances are, one of the max values are outside the image size. "
            msg += f"max={self.max.as_tuple} and image={image.GetSize()}"
            logger.warning(msg)

            new_max_coord = Coordinate(
                x=min(self.max.x, image.GetSize()[0]),
                y=min(self.max.y, image.GetSize()[1]),
                z=min(self.max.z, image.GetSize()[2]),
            )
            # bbox = BoundingBox(min=self.min, max=new_max_coord)
            # not a fan of mutating the object, but this adjustment
            # would be required for every use of the bounding box
            # so its better just to mutate this instance and let the fixed version be used everytime
            self.max = new_max_coord

            # re-run post_init to validate & update the size
            self.__post_init__()

            try:
                cropped_image = sitk.RegionOfInterest(
                    image,
                    self.size.as_tuple,
                    self.min.as_tuple,
                )
            except RuntimeError as e:
                msg = "Failed to crop the image to the bounding box.\n"
                msg += f"Image size: {image.GetSize()}. \n"
                msg += f"Bounding box: {self}\n"
                msg += "Failed retry crop to the largest possible region."
                msg += f" New bounding box: {self}"
                logger.error(msg)
                raise BoundingBoxError(msg) from e
        return cropped_image

    def crop_image_and_mask(
        self,
        image: sitk.Image,
        mask: sitk.Image,
    ) -> tuple[sitk.Image, sitk.Image]:
        """Crop the input image and mask to the bounding box.

        Parameters
        ----------
        image : sitk.Image
            The input image to crop.
        mask : sitk.Image
            The input mask to crop. Assumes they are aligned with the image.

        Returns
        -------
        tuple[sitk.Image, sitk.Image]
            The cropped image and mask.
        """
        return self.crop_image(image), self.crop_image(mask)

    def pad(self, padding: int) -> BoundingBox:
        """
        Expand the bounding box by a specified padding value in all directions.

        Parameters
        ----------
        padding : int
            The padding value to expand the bounding box.

        Returns
        -------
        BoundingBox
            The expanded bounding box.
        """
        if padding == 0:
            return self

        padded_min = Coordinate(
            x=self.min.x - padding,
            y=self.min.y - padding,
            z=self.min.z - padding,
        )
        padded_max = Coordinate(
            x=self.max.x + padding,
            y=self.max.y + padding,
            z=self.max.z + padding,
        )
        return BoundingBox(min=padded_min, max=padded_max)

    def expand_to_cube(self) -> BoundingBox:
        """Convert the bounding box to a cube by making the size equal along all dimensions.

        This is done by finding which dimension is the largest and then expanding the
        bounding box in the other dimensions to make it a cube.

        Returns
        -------
        BoundingBox
            The bounding box converted to a cube.
        """
        max_size = max(self.size.as_tuple)

        extra_x_half = (max_size - self.size.x) // 2
        extra_y_half = (max_size - self.size.y) // 2
        extra_z_half = (max_size - self.size.z) // 2

        min_coord = Coordinate(
            x=self.min.x - extra_x_half,
            y=self.min.y - extra_y_half,
            z=self.min.z - extra_z_half,
        )

        max_coord = Coordinate(
            x=self.max.x + extra_x_half,
            y=self.max.y + extra_y_half,
            z=self.max.z + extra_z_half,
        )

        # if any of the min values are negative, set them to 0, and add the difference to the max values
        if min_coord.x < 0:
            diff = abs(min_coord.x)
            min_coord.x = 0
            max_coord.x += diff

        if min_coord.y < 0:
            diff = abs(min_coord.y)
            min_coord.y = 0
            max_coord.y += diff

        if min_coord.z < 0:
            diff = abs(min_coord.z)
            min_coord.z = 0
            max_coord.z += diff

        return BoundingBox(min=min_coord, max=max_coord)



import time
from pathlib import Path
from typing import Dict, List, TypeVar

import tqdm
from pydicom import dcmread
from pydicom.errors import InvalidDicomError
from sqlalchemy.orm import Session

from imgtools.dicom.index.database import DatabaseHandler
from imgtools.dicom.index.models import (
    Image,
    Patient,
    Series,
    Study,
)
from imgtools.logging import logger


def _extract_metadata(file_path: Path, tags: List[str]) -> Dict[str, str]:
    """
    need to -> benchmark against `read_tags` function in `utils.py`
    """

    try:
        dicom = dcmread(file_path, specific_tags=tags, stop_before_pixels=True)
    except FileNotFoundError as fnfe:
        errmsg = f"File not found: {file_path}"
        raise FileNotFoundError(errmsg) from fnfe
    except InvalidDicomError as ide:
        errmsg = f"Invalid DICOM file: {file_path}"
        raise InvalidDicomError(errmsg) from ide
    except ValueError as ve:
        errmsg = f"Value error reading DICOM file: {file_path}"
        raise ValueError(errmsg) from ve

    return {tag: str(dicom.get(tag, "")) for tag in tags}


class DICOMInsertMixin:
    def _insert_patient(
        self, session: Session, metadata: Dict[str, str]
    ) -> Patient:
        patient = (
            session.query(Patient)
            .filter_by(PatientID=metadata["PatientID"])
            .first()
        )
        if not patient:
            patient = Patient.from_metadata(metadata)
            session.add(patient)
        return patient

    def _insert_study(
        self, session: Session, metadata: Dict[str, str]
    ) -> Study:
        study = (
            session.query(Study)
            .filter_by(StudyInstanceUID=metadata["StudyInstanceUID"])
            .first()
        )
        if not study:
            study = Study.from_metadata(metadata)
            session.add(study)
        return study

    def _insert_series(
        self, session: Session, metadata: Dict[str, str], file_path: Path
    ) -> Series:
        series = (
            session.query(Series)
            .filter_by(SeriesInstanceUID=metadata["SeriesInstanceUID"])
            .first()
        )
        if series:
            return series
        series = Series.from_metadata(metadata, file_path)
        session.add(series)
        return series

    def _insert_image(
        self, session: Session, metadata: Dict[str, str], file_path: Path
    ) -> Image:
        file = (
            session.query(Image)
            .filter_by(SOPInstanceUID=metadata["SOPInstanceUID"])
            .first()
        )
        if not file:
            file = Image.from_metadata(metadata, file_path)
            session.add(file)
        return file


class DICOMIndexer(DICOMInsertMixin):
    """
    A tool to index DICOM files and insert metadata into the database.
    """

    def __init__(self, db_handler: DatabaseHandler) -> None:
        """
        Initialize the DICOMIndexer.

        Parameters
        ----------
        db_handler : DatabaseHandler
                        An instance of DatabaseHandler for managing database operations.
        """
        self.db_handler = db_handler
        self.mytags = [
            "PatientID",
            "StudyInstanceUID",
            "SeriesInstanceUID",
            "Modality",
            "SOPInstanceUID",
        ]

    @property
    def existing_files(self) -> List[Path]:
        """
        Get a list of file paths already indexed in the database.

        Returns
        -------
        List[Path]
                        A list of file paths.
        """
        with self.db_handler.session() as session:
            return [
                Path(file.FilePath)
                for file in session.query(Image.FilePath).all()
            ]

    def build_index_from_files(self, files: List[Path]) -> None:
        """
        Build an index of DICOM files and insert metadata into the database.

        Parameters
        ----------
        files : List[Path]
                        List of file paths to index.
        """
        existing_files = self.existing_files
        remaining_files = list(set(files) - set(existing_files))
        logger.debug(f"{len(existing_files)} files already indexed.")
        logger.debug(
            f"{len(remaining_files)}/{len(files)} files left to index."
        )

        if not remaining_files:
            logger.info("All files are already indexed.")
            return
        start = time.time()
        try:
            with (
                self.db_handler.session() as session,
                tqdm.tqdm(
                    total=len(remaining_files),
                    initial=len(existing_files),
                    desc="Indexing DICOM files",
                ) as tqdm_files,
            ):
                for file_path in remaining_files:
                    metadata = _extract_metadata(file_path, self.mytags)
                    _ = self._insert_patient(session, metadata)
                    _ = self._insert_study(session, metadata)
                    series = self._insert_series(session, metadata, file_path)
                    image = self._insert_image(session, metadata, file_path)
                    series.images.append(image)
                    tqdm_files.update(1)
        except Exception as e:
            logger.exception(f"Error: {e}")
            raise e
        finally:
            logger.info(
                f"Indexing complete in {time.time() - start:.2f} seconds."
            )


T = TypeVar("T")  # Generic type for table models


class DICOMDatabaseInterface:
    """
    A class to interact with the DICOM database and perform high-level operations.
    """

    def __init__(self, db_handler: DatabaseHandler) -> None:
        """
        Initialize the database interface.

        Parameters
        ----------
        db_handler : DatabaseHandler
                An instance of DatabaseHandler for managing database operations.
        """
        self.session = db_handler.Session()

    @property
    def patients(self) -> List[Patient]:
        return self.session.query(Patient).all()

    @property
    def studies(self) -> List[Study]:
        return self.session.query(Study).all()

    @property
    def series(self) -> List[Series]:
        return self.session.query(Series).all()


# if __name__ == '__main__':
# 	from pathlib import Path

# 	from imgtools.dicom import find_dicoms


# 	directory = Path('/Users/bhklab/dev/radiomics/med-imagetools/data/nbia/images/unzipped')
# 	db_path = Path('dicom_index.sqlite')
# 	db_handler = DatabaseHandler(db_path=db_path)
# 	check_header = False
# 	extension = 'dcm'

# 	logger.info('Finding DICOM files...')
# 	dicom_files = find_dicoms(
# 		directory=directory,
# 		check_header=check_header,
# 		recursive=True,
# 		extension=extension,
# 	)
# 	logger.info(f'Found {len(dicom_files)} DICOM files.')

# 	indexer = DICOMIndexer(db_handler=db_handler)

# 	# Build index
# 	indexer.build_index_from_files(dicom_files[:1000])

# 	# Query the database
# 	db = DICOMDatabaseInterface(db_handler=db_handler)
# 	from rich import print  # noqa: A004

# 	start = time.time()
# 	print(db.patients[0])
# 	print('-' * 50)
# 	print(db.studies[0])
# 	print('-' * 50)
# 	print(db.series[0])
# 	print('-' * 50)
# 	print(db.series[0].images[0])
# 	print(f'Querying complete in {time.time() - start:.2f} seconds.')
# 	print(db.series[0].images[0].series.study.patient)



from __future__ import annotations

import glob
import json
import os
import pathlib
import re
from abc import ABC, abstractmethod
from collections import namedtuple
from typing import Optional, Union

import pandas as pd
import SimpleITK as sitk
from pydicom import dcmread

from imgtools.modules import PET, Dose, Scan, Segmentation, StructureSet
from imgtools.utils.dicomutils import get_modality_metadata


def read_image(path: str) -> sitk.Image:
    """Read an image from the specified file path using SimpleITK.

    Parameters
    ----------
    path : str
        The file path to the image.

    Returns
    -------
    sitk.Image
        The image read from the file.
    """
    return sitk.ReadImage(path)


def read_dicom_series(
    path: str,
    series_id: list[str] | None = None,
    recursive: bool = False,
    file_names: list[str] | None = None,
) -> sitk.Image:
    """Read DICOM series as SimpleITK Image.

    Parameters
    ----------
    path
       Path to directory containing the DICOM series.

    recursive, optional
       Whether to recursively parse the input directory when searching for
       DICOM series,

    series_id, optional
       Specifies the DICOM series to load if multiple series are present in
       the directory. If None and multiple series are present, loads the first
       series found.

    file_names, optional
        If there are multiple acquisitions/"subseries" for an individual series,
        use the provided list of file_names to set the ImageSeriesReader.

    Returns
    -------
    The loaded image.
    """
    reader = sitk.ImageSeriesReader()
    if file_names is None:
        # extract the names of the dicom files that are in the path variable, which is a directory
        file_names = reader.GetGDCMSeriesFileNames(
            path,
            seriesID=series_id if series_id else "",
            recursive=recursive,
        )

    reader.SetFileNames(file_names)

    # Configure the reader to load all of the DICOM tags (public+private):
    # By default tags are not loaded (saves time).
    # By default if tags are loaded, the private tags are not loaded.
    # We explicitly configure the reader to load tags, including the
    # private ones.
    reader.MetaDataDictionaryArrayUpdateOn()
    reader.LoadPrivateTagsOn()

    return reader.Execute()


def read_dicom_scan(
    path: str,
    series_id: list[str] | None = None,
    recursive: bool = False,
    file_names: list[str] | None = None,
) -> Scan:
    image = read_dicom_series(
        path,
        series_id=series_id,
        recursive=recursive,
        file_names=file_names,
    )
    return Scan(image, {})


def read_dicom_rtstruct(
    path: str,
    suppress_warnings: bool = False,
    roi_name_pattern: str | None = None,
) -> StructureSet:
    return StructureSet.from_dicom(
        path,
        suppress_warnings=suppress_warnings,
        roi_name_pattern=roi_name_pattern,
    )


def read_dicom_rtdose(path: str) -> Dose:
    return Dose.from_dicom(path=path)


def read_dicom_pet(path: str, series: Optional[str] = None) -> PET:
    return PET.from_dicom(path=path, series_id=series, pet_image_type="SUV")


def read_dicom_seg(
    path: str, meta: dict, series: Optional[str] = None
) -> Segmentation:
    seg_img = read_dicom_series(path, series)
    return Segmentation.from_dicom(seg_img, meta)


auto_dicom_result = Union[Scan, PET, StructureSet, Dose, Segmentation]


def read_dicom_auto(
    path: str, series=None, file_names=None
) -> auto_dicom_result:
    dcms = (
        list(pathlib.Path(path).rglob("*.dcm"))
        if not path.endswith(".dcm")
        else [pathlib.Path(path)]
    )

    for dcm_path in dcms:
        dcm = dcm_path.as_posix()
        meta = dcmread(dcm, stop_before_pixels=True)
        if meta.SeriesInstanceUID != series and series is not None:
            continue

        modality = meta.Modality

        match modality:
            case "CT" | "MR":
                obj = read_dicom_scan(path, series, file_names=file_names)
            case "PT":
                obj = read_dicom_pet(path, series)
            case "RTSTRUCT":
                obj = read_dicom_rtstruct(dcm)
            case "RTDOSE":
                obj = read_dicom_rtdose(dcm)
            case "SEG":
                obj = read_dicom_seg(path, meta, series)
            case _:
                errmsg = (
                    f"Modality {modality} not supported in read_dicom_auto."
                )
                raise NotImplementedError(errmsg)

        obj.metadata.update(get_modality_metadata(meta, modality))
        return obj


# ruff: noqa
class BaseLoader(ABC):
    @abstractmethod
    def __getitem__(self, subject_id):
        pass

    def __len__(self) -> int:
        return len(self.keys())

    @abstractmethod
    def keys(self):
        pass

    def items(self):
        return ((k, self[k]) for k in self.keys())

    def values(self):
        return (self[k] for k in self.keys())

    def get(self, subject_id, default=None):
        try:
            return self[subject_id]
        except KeyError:
            return default


class ImageTreeLoader(BaseLoader):
    def __init__(
        self,
        json_path,
        csv_path_or_dataframe,
        col_names=None,
        study_names=None,
        series_names=None,
        subseries_names=None,
        id_column=None,
        expand_paths=False,
        readers=None,
    ) -> None:
        if subseries_names is None:
            subseries_names = []
        if series_names is None:
            series_names = []
        if study_names is None:
            study_names = []
        if col_names is None:
            col_names = []
        if readers is None:
            readers = [
                read_image
            ]  # no mutable defaults https://florimond.dev/en/posts/2018/08/python-mutable-defaults-are-the-source-of-all-evil/

        self.expand_paths = expand_paths
        self.readers = readers
        self.colnames = col_names
        self.studynames = study_names
        self.seriesnames = series_names
        self.subseriesnames = subseries_names

        if isinstance(csv_path_or_dataframe, str):
            if id_column is not None and id_column not in self.colnames:
                self.colnames.append(id_column)
            self.paths = pd.read_csv(
                csv_path_or_dataframe, index_col=id_column
            )
        elif isinstance(csv_path_or_dataframe, pd.DataFrame):
            self.paths = csv_path_or_dataframe
            if id_column:
                self.paths = self.paths.set_index(id_column)
            if len(self.colnames) == 0:
                self.colnames = self.paths.columns
        else:
            msg = f"Expected a path to csv file or pd.DataFrame, not {type(csv_path_or_dataframe)}."
            raise ValueError(msg)

        if isinstance(json_path, str):
            with open(json_path, "r") as f:
                self.tree = json.load(f)
        else:
            msg = f"Expected a path to a json file, not {type(json_path)}."
            raise ValueError(msg)

        if not isinstance(readers, list):
            readers = [readers] * len(self.colnames)

        self.output_tuple = namedtuple("Output", self.colnames)

    def __getitem__(self, subject_id):
        row = self.paths.loc[subject_id]
        paths = {col: row[col] for col in self.colnames}
        study = {col: row[col] for col in self.studynames}
        series = {col: row[col] for col in self.seriesnames}
        subseries = {col: row[col] for col in self.subseriesnames}
        paths = {k: v if pd.notna(v) else None for k, v in paths.items()}

        if self.expand_paths:
            # paths = {col: glob.glob(path)[0] for col, path in paths.items()}
            paths = {
                col: glob.glob(path)[0] if pd.notna(path) else None
                for col, path in paths.items()
            }

        for i, (col, path) in enumerate(paths.items()):
            files = self.tree[subject_id][
                study["study_" + ("_").join(col.split("_")[1:])]
            ][series["series_" + ("_").join(col.split("_")[1:])]][
                subseries["subseries_" + ("_").join(col.split("_")[1:])]
            ]
            self.readers[i](
                path, series["series_" + ("_").join(col.split("_")[1:])]
            )
        outputs = {
            col: self.readers[i](
                path,
                series["series_" + ("_").join(col.split("_")[1:])],
                file_names=files,
            )
            for i, (col, path) in enumerate(paths.items())
        }
        return self.output_tuple(**outputs)

    def keys(self):
        return list(self.paths.index)

    def items(self):
        return ((k, self[k]) for k in self.keys())


class ImageCSVLoader(BaseLoader):
    def __init__(
        self,
        csv_path_or_dataframe,
        colnames=None,
        seriesnames=None,
        id_column=None,
        expand_paths=False,
        readers=None,
    ) -> None:
        if seriesnames is None:
            seriesnames = []
        if colnames is None:
            colnames = []
        if readers is None:
            readers = [
                read_image
            ]  # no mutable defaults https://florimond.dev/en/posts/2018/08/python-mutable-defaults-are-the-source-of-all-evil/

        self.expand_paths = expand_paths
        self.readers = readers

        self.colnames = colnames
        self.seriesnames = seriesnames
        if isinstance(csv_path_or_dataframe, str):
            if id_column is not None and id_column not in colnames:
                colnames.append(id_column)
            self.paths = pd.read_csv(
                csv_path_or_dataframe, index_col=id_column
            )
        elif isinstance(csv_path_or_dataframe, pd.DataFrame):
            self.paths = csv_path_or_dataframe
            if id_column:
                self.paths = self.paths.set_index(id_column)
            if len(self.colnames) == 0:
                self.colnames = self.paths.columns
        else:
            msg = f"Expected a path to csv file or pd.DataFrame, not {type(csv_path_or_dataframe)}."
            raise ValueError(msg)

        if not isinstance(readers, list):
            readers = [readers] * len(self.colnames)

        self.output_tuple = namedtuple("Output", self.colnames)

    def __getitem__(self, subject_id):
        row = self.paths.loc[subject_id]
        paths = {col: row[col] for col in self.colnames}
        series = {col: row[col] for col in self.seriesnames}
        paths = {k: v if pd.notna(v) else None for k, v in paths.items()}
        if self.expand_paths:
            # paths = {col: glob.glob(path)[0] for col, path in paths.items()}
            paths = {
                col: glob.glob(path)[0] if pd.notna(path) else None
                for col, path in paths.items()
            }

        outputs = {
            col: self.readers[i](
                path, series["series_" + ("_").join(col.split("_")[1:])]
            )
            for i, (col, path) in enumerate(paths.items())
        }
        return self.output_tuple(**outputs)

    def keys(self):
        return list(self.paths.index)

    def items(self):
        return ((k, self[k]) for k in self.keys())


class ImageFileLoader(BaseLoader):
    def __init__(
        self,
        root_directory,
        get_subject_id_from="filename",
        subdir_path=None,
        exclude_paths=None,
        reader=None,
    ) -> None:
        if exclude_paths is None:
            exclude_paths = []
        if reader is None:
            reader = read_image  # no mutable defaults https://florimond.dev/en/posts/2018/08/python-mutable-defaults-are-the-source-of-all-evil/

        self.root_directory = root_directory
        self.get_subject_id_from = get_subject_id_from
        self.subdir_path = subdir_path
        self.exclude_paths = []
        for path in exclude_paths:
            if not path.startswith(self.root_directory):
                full_paths = glob.glob(
                    pathlib.Path(root_directory, path).as_posix()
                )
                self.exclude_paths.extend(full_paths)
            else:
                full_path = path
                self.exclude_paths.append(full_path)
        self.reader = reader

        self.paths = self._generate_paths()

    def _generate_paths(self):
        paths = {}
        for f in os.scandir(self.root_directory):
            if f.path in self.exclude_paths:
                continue
            subject_dir_path = f.path
            if self.subdir_path:
                full_path = pathlib.Path(
                    subject_dir_path, self.subdir_path
                ).as_posix()
            else:
                full_path = subject_dir_path
            try:
                full_path = glob.glob(full_path)[0]
            except IndexError:
                continue
            if os.path.isdir(full_path):
                full_path = pathlib.Path(full_path, "").as_posix()
            subject_dir_name = os.path.basename(
                os.path.normpath(subject_dir_path)
            )
            subject_id = self._extract_subject_id_from_path(
                full_path, subject_dir_name
            )
            paths[subject_id] = full_path
        return paths

    def _extract_subject_id_from_path(self, full_path, subject_dir_name):
        filename, _ = os.path.splitext(os.path.basename(full_path))
        if isinstance(self.get_subject_id_from, str):
            if self.get_subject_id_from == "filename":
                subject_id = filename
            elif self.get_subject_id_from == "subject_directory":
                subject_id = subject_dir_name
            else:
                subject_id = re.search(self.get_subject_id_from, full_path)[0]
        else:
            return self.get_subject_id_from(
                full_path, filename, subject_dir_name
            )
        return subject_id

    def __getitem__(self, subject_id):
        path = self.paths[subject_id]
        return self.reader(path)

    def keys(self):
        return self.paths.keys()




import os
import shutil
import sys
from pathlib import Path

import pytest
from typing import Tuple


def _win32_longpath(path):
    """
    Helper function to add the long path prefix for Windows, so that shutil.copytree
     won't fail while working with paths with 255+ chars.
    """
    if sys.platform == "win32":
        # The use of os.path.normpath here is necessary since "the "\\?\" prefix
        # to a path string tells the Windows APIs to disable all string parsing
        # and to send the string that follows it straight to the file system".
        # (See https://docs.microsoft.com/pt-br/windows/desktop/FileIO/naming-a-file)
        normalized = os.path.normpath(path)
        if not normalized.startswith("\\\\?\\"):
            is_unc = normalized.startswith("\\\\")
            # see https://en.wikipedia.org/wiki/Path_(computing)#Universal_Naming_Convention # noqa: E501
            if (
                is_unc
            ):  # then we need to insert an additional "UNC\" to the longpath prefix
                normalized = normalized.replace("\\\\", "\\\\?\\UNC\\")
            else:
                normalized = "\\\\?\\" + normalized
        return normalized
    else:
        return path


@pytest.fixture(scope="function")
def legacy_test_data(
    quebec_paths: Tuple[str, str, str, str],
    request: pytest.FixtureRequest,
    tmp_path: Path
) -> Tuple[str, str, str, str]:
    input_path, output_path, crawl_path, edge_path = quebec_paths

    assert Path(input_path).exists(), "Dataset not found"
    # idea here is to copy the data to a temporary directory
    # so that the tests can be run in isolation

    fixture_name = request.fspath.purebasename

    result = tmp_path / fixture_name

    shutil.copytree(_win32_longpath(input_path), _win32_longpath(result), symlinks=True, dirs_exist_ok=True)

    # remake the other paths

    crawl_path = result.parent / ".imgtools" / f"imgtools_{fixture_name}.csv"
    edge_path = result.parent / ".imgtools" / f"imgtools_{fixture_name}_edges.csv"
    output_path = result / "temp"

    output_path.mkdir(parents=True, exist_ok=True)
    crawl_path.parent.mkdir(parents=True, exist_ok=True)
    edge_path.parent.mkdir(parents=True, exist_ok=True)

    return result.as_posix(), output_path.as_posix(), crawl_path.as_posix(), edge_path.as_posix()



from pathlib import Path
from typing import List

import pytest

from imgtools.datasets.github_helper import (
    GitHubRelease,
    GitHubReleaseAsset,
    MedImageTestData,
)


@pytest.fixture
def med_image_test_data() -> MedImageTestData:
    manager = MedImageTestData()
    _ = manager.get_latest_release()
    return manager


def test_get_latest_release_and_extract(med_image_test_data: MedImageTestData, tmp_path: Path) -> None:
    # Test get_latest_release
    latest_release: GitHubRelease = med_image_test_data.get_latest_release()
    assert latest_release is not None
    assert latest_release.tag_name is not None
    assert latest_release.name is not None
    assert isinstance(latest_release.tag_name, str)
    assert isinstance(latest_release.name, str)

def test_downloaded_data_is_sound(download_all_test_data: dict[str, Path])-> None:
    assert len(download_all_test_data) > 5 
    datasets_of_interest = [
        "NSCLC-Radiomics",
        "NSCLC_Radiogenomics",
        "Vestibular-Schwannoma-SEG",
    ]

    assert all(
        dataset in download_all_test_data.keys() for dataset in datasets_of_interest
    )
    assert all(
        download_all_test_data[dataset].exists() for dataset in datasets_of_interest
    )



from functools import reduce
from pathlib import Path
from typing import Callable, List, Tuple

import numpy as np
import pandas as pd

from imgtools.logging import logger


class DataGraph:
    """
    This class given the crawled dataset in the form of CSV file, deals with forming a graph on the full dataset, taking advantage of connections between different modalities. Based
    on these connections, an edge table is made. This class also supports querying and for a given query, returns the file locations of the user-specified sub-dataset from the full dataset.
    The graph is made based on the references made by different DICOMS. Different connections are given different edge type values, to make parsing easier. The edge types are as follows:-
    1) edge_type:0 RTDOSE(key:ref_rt) -> RTSTRUCT(pair: series/instance)
    2) edge_type:1 RTDOSE(key:ref_ct) -> CT(pair: series)
    3) edge_type:2 RTSTRUCT(key:ref_ct) -> CT(pair: series)
    4) edge_type:3 RTSTRUCT(key:ref_ct) -> PT(pair: series)
    5) edge_type:4 CT(key:study) -> PT(pair: study)
    6) edge_type:5 RTDOSE(key: ref_pl) -> RTPLAN(pair: instance)
    7) edge_type:6 RTPLAN(key: ref_rs) -> RTSTRUCT(pair: series/instance)
    8) edge_type:7 SEG(key:ref_ct) -> CT(pair: series)

    Once the edge table is formed, one can query on the graph to get the desired results. For uniformity, the supported query is list of modalities to consider
    For ex:
    query = ["CT","RTDOSE","RTSTRUCT","PT], will return interconnected studies containing the listed DICOM modalities. The interconnected studies for example may look like
    (RTDOSE->RTSTRUCT->CT<-PT<-RTSTRUCT)
    """

    def __init__(
        self,
        path_crawl: str | Path,
        edge_path: str | Path = "./patient_id_full_edges.csv",
        visualize: bool = False,
        update: bool = False,
    ) -> None:
        """
        Parameters
        ----------
        path_crawl: str | Path
            The csv returned by the crawler

        edge_path: str | Path, default = "./patient_id_full_edges.csv"
            This path denotes where the graph in the form of edge table is stored or to be stored

        visualize: bool, default = False
            Whether to generate graph visualization using Pyviz

        update: bool, default = False
            Whether to force update existing edge table
        """
        self.df = pd.read_csv(path_crawl, index_col=0)
        self.edge_path = Path(edge_path)
        self.df_new = None

        if not self.edge_path.exists():
            logger.info(
                "Edge table not present. Forming the edge table based on the crawl data..."
            )
            self.form_graph()
        elif not update:
            logger.info("Edge table is already present. Loading the data...")
            self.df_edges = pd.read_csv(self.edge_path)
        else:
            logger.info("Edge table present, but force updating...")
            self.form_graph()
        if visualize:
            self.visualize_graph()

    def form_graph(self) -> None:
        """
        Forms edge table based on the crawled data
        """
        # Enforce string type to all columns to prevent dtype merge errors for empty columns
        self.df = self.df.astype(str)

        # Get reference_rs information from RTDOSE-RTPLAN connections
        df_filter = pd.merge(
            self.df,
            self.df[["instance_uid", "reference_rs"]],
            left_on="reference_pl",
            right_on="instance_uid",
            how="left",
        )

        df_filter.loc[
            (df_filter.reference_rs_x.isna())
            & (~df_filter.reference_rs_y.isna()),
            "reference_rs_x",
        ] = df_filter.loc[
            (df_filter.reference_rs_x.isna())
            & (~df_filter.reference_rs_y.isna()),
            "reference_rs_y",
        ].values
        df_filter.drop(
            columns=["reference_rs_y", "instance_uid_y"], inplace=True
        )
        df_filter.rename(
            columns={
                "reference_rs_x": "reference_rs",
                "instance_uid_x": "instance_uid",
            },
            inplace=True,
        )

        # Remove entries with no RTDOSE reference, for extra check, such cases are most probably removed in the earlier step
        df_filter = df_filter.loc[
            ~(
                (df_filter["modality"] == "RTDOSE")
                & (df_filter["reference_ct"].isna())
                & (df_filter["reference_rs"].isna())
            )
        ]

        self.df_edges = self._form_edges(self.df)
        self.df_edges.loc[self.df_edges.study_x.isna(), "study_x"] = (
            self.df_edges.loc[self.df_edges.study_x.isna(), "study"]
        )
        self.df_edges.drop(
            columns=[
                "study_y",
                "patient_ID_y",
                "series_description_y",
                "study_description_y",
                "study",
            ],
            inplace=True,
        )
        self.df_edges.sort_values(by="patient_ID_x", ascending=True)

        logger.info(f"Saving edge table in {self.edge_path}")
        self.df_edges.to_csv(self.edge_path, index=False)

    def visualize_graph(self) -> None:
        """
        Generates visualization using Pyviz, a wrapper around visJS. The visualization can be found at datanet.html
        """
        try:
            from pyvis.network import Network  # type: ignore
        except ImportError:
            import sys

            logger.error("Please install `pyvis` to visualize the graph")
            sys.exit(1)

        logger.info("Generating visualizations...")
        data_net = Network(
            height="100%", width="100%", bgcolor="#222222", font_color="white"
        )

        source_series = self.df_edges["series_y"]
        target_series = self.df_edges["series_x"]
        source_modality = self.df_edges["modality_y"]
        target_modality = self.df_edges["modality_x"]
        patient_id = self.df_edges["patient_ID_x"]
        reference_ct = self.df_edges["reference_ct_y"]
        reference_rs = self.df_edges["reference_rs_y"]

        data_zip = zip(
            source_series,
            target_series,
            source_modality,
            target_modality,
            patient_id,
            reference_ct,
            reference_rs,
        )

        for src_s, targ_s, src_m, targ_m, p_id, ref_ct, ref_rs in data_zip:
            data_net.add_node(src_s, src_m, title=src_m, group=p_id)
            data_net.add_node(targ_s, targ_m, title=targ_m, group=p_id)
            data_net.add_edge(src_s, targ_s)

            node = data_net.get_node(src_s)
            node["title"] = (
                "<br>Patient_id: {}<br>Series: {}<br>reference_ct: {}<br>reference_rs: {}".format(
                    p_id, src_s, ref_ct, ref_rs
                )
            )

            node = data_net.get_node(targ_s)
            node["title"] = (
                "<br>Patient_id: {}<br>Series: {}<br>reference_ct: {}<br>reference_rs: {}".format(
                    p_id, targ_s, ref_ct, ref_rs
                )
            )

        neigbour_map = data_net.get_adj_list()
        for node in data_net.nodes:
            node["title"] += "<br>Number of connections: {}".format(
                len(neigbour_map[node["id"]])
            )
            node["value"] = len(neigbour_map[node["id"]])

        vis_path = self.edge_path.parent / "datanet.html"
        logger.info(f"Saving HTML of visualization at {vis_path}")
        data_net.show(vis_path)

    def _form_edges(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        For a given study id forms edge table
        """

        df_list = []

        # Split into each modality
        plan = df[df["modality"] == "RTPLAN"]
        dose = df[df["modality"] == "RTDOSE"]
        struct = df[df["modality"] == "RTSTRUCT"]
        seg = df[df["modality"] == "SEG"]
        ct = df[df["modality"] == "CT"]
        mr = df[df["modality"] == "MR"]
        pet = df[df["modality"] == "PT"]

        for edge in range(8):
            # FORMS RTDOSE->RTSTRUCT, can be formed on both series and instance uid
            if edge == 0:
                df_combined = pd.concat(
                    [
                        pd.merge(
                            struct,
                            dose,
                            left_on="instance_uid",
                            right_on="reference_rs",
                        ),
                        pd.merge(
                            struct,
                            dose,
                            left_on="series",
                            right_on="reference_rs",
                        ),
                    ]
                ).drop_duplicates(
                    subset=["instance_uid_x"]
                )  # drop_duplicates for cases where both series and instance_uid are the same for struct

            # FORMS RTDOSE->CT
            elif edge == 1:
                df_combined = pd.merge(
                    ct, dose, left_on="series", right_on="reference_ct"
                )

            # FORMS RTSTRUCT->CT on ref_ct to series
            elif edge == 2:
                df_combined = pd.concat(
                    [
                        pd.merge(
                            ct,
                            struct,
                            left_on="series",
                            right_on="reference_ct",
                        ),
                        pd.merge(
                            mr,
                            struct,
                            left_on="series",
                            right_on="reference_ct",
                        ),
                    ]
                )

            # FORMS RTSTRUCT->PET on ref_ct to series
            elif edge == 3:
                df_combined = pd.merge(
                    pet, struct, left_on="series", right_on="reference_ct"
                )

            # FORMS PET->CT on study
            elif edge == 4:
                df_combined = pd.merge(
                    ct, pet, left_on="study", right_on="study"
                )

            # FORMS RTPLAN->RTDOSE on ref_pl
            elif edge == 5:
                df_combined = pd.merge(
                    plan, dose, left_on="instance_uid", right_on="reference_pl"
                )

            # FORMS RTSTRUCT->RTPLAN on ref_rs
            elif edge == 6:
                df_combined = pd.merge(
                    struct,
                    plan,
                    left_on="instance_uid",
                    right_on="reference_rs",
                )

            # FORMS SEG->CT/MR
            elif edge == 7:
                df_combined = pd.concat(
                    [
                        pd.merge(
                            ct, seg, left_on="series", right_on="reference_ct"
                        ),
                        pd.merge(
                            mr, seg, left_on="series", right_on="reference_ct"
                        ),
                    ]
                )

            df_combined["edge_type"] = edge
            df_list.append(df_combined)

        df_edges = pd.concat(df_list, axis=0, ignore_index=True)
        return df_edges

    def parser(self, query_string: str) -> pd.DataFrame:  # noqa
        """
        For a given query string(Check the documentation), returns the dataframe consisting of
        two columns namely modality and folder location of the connected nodes
        Parameters
        ----------
        query_string: str
            Query string based on which dataset will be formed

        Query ideas:
        There are four basic supported modalities are RTDOSE, RTSTRUCT, CT, PT, MR
        The options are, the string can be in any order:
        1) RTDOSE
        2) RTSTRUCT
        3) CT
        4) PT
        5) PT,RTSTRUCT
        6) CT,PT
        7) CT,RTSTRUCT
        8) CT,RTDOSE
        9) RTDOSE,RTSTRUCT,CT
        10) RTDOSE,CT,PT
        11) RTSTRUCT,CT,PT
        12) RTDOSE,RTSTRUCT,CT,PT
        """

        # Supported modalities and edge definitions
        supported_modalities = ["RTDOSE", "RTSTRUCT", "CT", "PT", "MR", "SEG"]
        edge_definitions = {
            "RTSTRUCT,RTDOSE": 0,
            "CT,RTDOSE": 1,
            "CT,RTSTRUCT": 2,
            "MR,RTSTRUCT": 2,
            "PET,RTSTRUCT": 3,
            "CT,PT": 4,
            "RTPLAN,RTDOSE": 5,
            "RTPLAN,RTSTRUCT": 6,
            "CT,SEG": 7,
            "MR,SEG": 7,
        }

        self.queried_modalities = query_string.split(",")

        # Handle single-modality queries
        if query_string in supported_modalities:
            final_df = self.df.loc[
                self.df["modality"] == query_string,
                ["study", "patient_ID", "series", "folder", "subseries"],
            ]
            final_df.rename(
                columns={
                    "series": f"series_{query_string}",
                    "study": f"study_{query_string}",
                    "folder": f"folder_{query_string}",
                    "subseries": f"subseries_{query_string}",
                },
                inplace=True,
            )
        # Handle pair-modality queries
        elif len(self.queried_modalities) == 2:
            # Determine the valid query by checking the original and reversed modality pairs in edge definitions
            valid_query = (
                query_string
                if query_string in edge_definitions
                else ",".join(self.queried_modalities[::-1])
            )
            edge_type = edge_definitions.get(valid_query)

            if edge_type is None:
                raise ValueError("Invalid Query. Select valid pairs.")

            # For cases such as the CT-RTSTRUCT and CT-RTDOSE, there exists multiple pathways due to which just searching on the edgetype gives wrong results
            if edge_type == 0:
                # Search for subgraphs with edges (0 or (5 and 6)) or (1 and 2)
                edge_condition = lambda row: (
                    (
                        "0" in row["edge_type"]
                        or (
                            "5" in row["edge_type"] and "6" in row["edge_type"]
                        )
                    )
                    or ("1" in row["edge_type"] and "2" in row["edge_type"])
                )
                mod = [
                    i for i in self.queried_modalities if i in ["CT", "MR"]
                ][
                    0
                ]  # making folder_mod CT/MR agnostic <-- still needs testing
                final_df = self.graph_query(
                    edge_condition, [0, 1, 2], f"folder_{mod}"
                )
            elif edge_type == 1:
                # Search for subgraphs with edges 1 or ((0 or (5 and 6)) and 2)
                edge_condition = lambda row: (
                    "1" in row["edge_type"]
                    or (
                        (
                            "0" in row["edge_type"]
                            or (
                                "5" in row["edge_type"]
                                and "6" in row["edge_type"]
                            )
                        )
                        and "2" in row["edge_type"]
                    )
                )
                final_df = self.graph_query(
                    edge_condition, [0, 1, 2], "RTSTRUCT"
                )
            elif edge_type == 2:
                # Search for subgraphs with edges 2 or ((0 or (5 and 6)) and 1)
                edge_condition = lambda row: (
                    "2" in row["edge_type"]
                    or (
                        (
                            "0" in row["edge_type"]
                            or (
                                "5" in row["edge_type"]
                                and "6" in row["edge_type"]
                            )
                        )
                        and "1" in row["edge_type"]
                    )
                )
                final_df = self.graph_query(
                    edge_condition, [0, 1, 2], "RTDOSE"
                )
            elif edge_type == 7:  # SEG->CT/MR
                # keep final_df as is
                final_df = self.df_edges.loc[
                    self.df_edges.edge_type == edge_type
                ].copy()
                node_dest, node_origin = valid_query.split(",")
                final_df.rename(
                    columns={
                        "study_x": "study",
                        "patient_ID_x": "patient_ID",
                        "series_x": f"series_{node_dest}",
                        "series_y": f"series_{node_origin}",
                        "folder_x": f"folder_{node_dest}",
                        "folder_y": f"folder_{node_origin}",
                        "subseries_x": f"subseries_{node_dest}",
                        "subseries_y": f"subseries_{node_origin}",
                    },
                    inplace=True,
                )
            else:
                final_df = self.df_edges.loc[
                    self.df_edges.edge_type == edge_type,
                    [
                        "study",
                        "patient_ID_x",
                        "study_x",
                        "study_y",
                        "series_x",
                        "folder_x",
                        "series_y",
                        "folder_y",
                        "subseries_x",
                        "subseries_y",
                    ],
                ]
                node_dest = valid_query.split(",")[0]
                node_origin = valid_query.split(",")[1]
                final_df.rename(
                    columns={
                        "study": "study",
                        "patient_ID_x": "patient_ID",
                        "series_x": f"series_{node_dest}",
                        "series_y": f"series_{node_origin}",
                        "study_x": f"study_{node_dest}",
                        "study_y": f"study_{node_origin}",
                        "folder_x": f"folder_{node_dest}",
                        "folder_y": f"folder_{node_origin}",
                        "subseries_x": f"subseries_{node_dest}",
                        "subseries_y": f"subseries_{node_origin}",
                    },
                    inplace=True,
                )
        # Handle combinations of modality
        elif len(self.queried_modalities) > 2:
            bads = ["RTPLAN"]
            # CT/MR,RTSTRUCT,RTDOSE
            if (
                (("CT" in query_string) or ("MR" in query_string))
                & ("RTSTRUCT" in query_string)
                & ("RTDOSE" in query_string)
                & ("PT" not in query_string)
            ):
                # Fetch the required data. Checks whether each study has edge 2 and (1 or (0 or (5 and 6)))
                edge_condition = lambda row: (
                    "2" in row["edge_type"]
                    and (
                        "1" in row["edge_type"]
                        or "0" in row["edge_type"]
                        or (
                            "5" in row["edge_type"] and "6" in row["edge_type"]
                        )
                    )
                )
                edge_list = [0, 1, 2, 5, 6]
            # CT/MR,RTSTRUCT,RTDOSE,PT
            elif (
                (("CT" in query_string) or ("MR" in query_string))
                & ("RTSTRUCT" in query_string)
                & ("RTDOSE" in query_string)
                & ("PT" in query_string)
            ):
                # Fetch the required data. Checks whether each study has edge (1 or (0 or (5 and 6))) and 2,3,4
                edge_condition = lambda row: (
                    (
                        "1" in row["edge_type"]
                        or "0" in row["edge_type"]
                        or (
                            "5" in row["edge_type"] and "6" in row["edge_type"]
                        )
                    )
                    and "2" in row["edge_type"]
                    and "3" in row["edge_type"]
                    and "4" in row["edge_type"]
                )
                edge_list = [0, 1, 2, 3, 4]
            # CT/MR,RTSTRUCT,PT
            elif (
                (("CT" in query_string) or ("MR" in query_string))
                & ("RTSTRUCT" in query_string)
                & ("PT" in query_string)
                & ("RTDOSE" not in query_string)
            ):
                # Fetch the required data. Checks whether each study has edge 2,3,4
                edge_condition = (
                    lambda row: "2" in row["edge_type"]
                    and "3" in row["edge_type"]
                    and "4" in row["edge_type"]
                )
                edge_list = [2, 3, 4]
            # CT/MR,RTDOSE,PT
            elif (
                (("CT" in query_string) or ("MR" in query_string))
                & ("RTSTRUCT" not in query_string)
                & ("PT" in query_string)
                & ("RTDOSE" in query_string)
            ):
                # Fetch the required data. Checks whether each study has edge 4 and (1 or (2 and (0 or (5 and 6)))). Remove RTSTRUCT later
                edge_condition = lambda row: (
                    "4" in row["edge_type"]
                    and (
                        "1" in row["edge_type"]
                        or (
                            "2" in row["edge_type"]
                            and (
                                "0" in row["edge_type"]
                                or (
                                    "5" in row["edge_type"]
                                    and "6" in row["edge_type"]
                                )
                            )
                        )
                    )
                )
                edge_list = [0, 1, 2, 4, 5, 6]
                bads.append("RTSTRUCT")
            else:
                raise ValueError("Please enter the correct query")

            final_df = self.graph_query(edge_condition, edge_list, bads)
        else:
            raise ValueError("Please enter the correct query")

        # Reset index and set index to patient_ID
        # drop means the old index is removed
        final_df.reset_index(drop=True, inplace=True)
        # Set the index to the format of "{index}_{patient_ID}"
        # i.e "0_JohnDoe-34" "1_JohnDoe-34"
        # This helps separate the same patient with multiple edges / connections
        # aka if multiple Studies are present
        final_df["index_chng"] = (
            final_df.index.astype(str)
            + "_"
            + final_df["patient_ID"].astype(str)
        )
        final_df.set_index("index_chng", inplace=True)
        final_df.rename_axis(None, inplace=True)

        # Change relative paths to absolute paths
        for col in final_df.columns:
            if col.startswith("folder"):
                final_df[col] = final_df[col].apply(
                    lambda x: (self.edge_path.parent.parent / x)
                    .resolve()
                    .as_posix()
                    if isinstance(x, str)
                    else x
                )  # input folder joined with the rel path

        return final_df

    def graph_query(
        self,
        edge_condition: Callable,
        edge_list: List[int],
        change_df: List[str],
        return_components: bool = False,
        remove_less_comp: bool = True,
    ) -> pd.DataFrame | list:
        """
        Based on the regex forms the final dataframe. You can
        query the edge table based on the regex to get the
        subgraph in which the queried edges will be present.

        The components are process further to get the final
        dataframe of the required modalities.

        Parameters
        ----------
        edge_condition: Callable
            To search the string in edge_type column of self.df_new which is aggregate of all the edges in a single study

        edge_list: List[int]
            The list of edges that should be returned in the subgraph

        change_df: List[str]
            Use only when you want to remove columns containing that string

        return_components: bool, default = False
            True to return the dictionary of the componets present with the condition present in the regex

        remove_less_comp: bool, default = True
            False when you want to keep components with modalities less than the modalitiy listed in the query
        """
        if self.df_new is None:
            self._form_agg()  # Form aggregates

        relevant_study_id = self.df_new.loc[
            self.df_new.apply(edge_condition, axis=1), "study_x"
        ].unique()

        # Based on the correct study ids, fetches the relevant edges
        df_processed = self.df_edges.loc[
            self.df_edges.study_x.isin(relevant_study_id)
            & (self.df_edges.edge_type.isin(edge_list))
        ]

        # The components are deleted if it has less number of nodes than the passed modalities, change this so as to alter that condition
        final_df = self._get_df(
            df_processed, relevant_study_id, remove_less_comp
        )

        # Removing columns
        for bad in change_df:
            # Find columns with change_df string present
            col_ids = [
                cols
                for cols in list(final_df.columns)[1:]
                if bad != cols.split("_")[1]
            ]
            final_df = final_df[[*list(final_df.columns)[:1], *col_ids]]

        if return_components:
            return self.final_dict
        else:
            return final_df

    def _form_agg(self) -> None:
        """
        Form aggregates for easier parsing, gets the edge types for each study and aggregates as a string. This way one can do regex based on what type of subgraph the user wants
        """

        def list_edges(series) -> str:  # noqa
            return reduce(lambda x, y: str(x) + str(y), series)

        self.df_edges["edge_type_str"] = self.df_edges["edge_type"].astype(str)
        self.df_new = self.df_edges.groupby("study_x").agg(
            {"edge_type_str": list_edges}
        )
        self.df_new.reset_index(level=0, inplace=True)
        self.df_new["edge_type"] = self.df_new["edge_type_str"]

    def _get_df(
        self,
        df_edges_processed: pd.DataFrame,
        rel_studyids: np.ndarray,
        remove_less_comp: bool = True,
    ) -> pd.DataFrame:
        """
        Assumption
        ----------
        The components are determined based on the unique CTs.
        Please ensure the data conforms to this case. Based on
        our preliminary analysis, there are no cases where CT
        and PT are present but are disconnected.

        Hence this assumption should hold for most of the cases
        This function returns dataframe consisting of folder
        location and modality for subgraphs

        Parameters
        ----------
        df_edges_processed: pd.Dataframe
            Dataframe processed containing only the desired edges from the full graph

        rel_studyids: np.ndarray
            Relevant study ids to process(This operation is a bit costly
            so better not to perform on full graph for maximum performance)

        remove_less_comp: bool, default = True
            True for removing components with less number of edges than the query

        Changelog
        ---------
        * June 14th, 2022: Changing from studyID-based to sample-based for loop
        * Oct 11th, 2022: Reverted to studyID-based loop + improved readability and make CT,RTSTRUCT,RTDOSE mode pass tests
        """
        self.final_dict = []
        final_df = []
        desired_modalities = set(self.queried_modalities)

        # Determine the number of components
        for _, study in enumerate(rel_studyids):
            df_temp = df_edges_processed.loc[
                df_edges_processed["study_x"] == study
            ]

            ct_locs = df_temp.loc[df_temp.modality_x.isin(["CT", "MR"])]
            ct_series = ct_locs.series_x.unique()

            comp, save_folder_comp = [], []

            # Initialization - For each component intialize a dictionary with the CTs and their connections
            for ct in ct_series:
                df_connections = ct_locs.loc[ct_locs.series_x == ct]
                row = (
                    df_connections.iloc[0]
                    if len(df_connections) > 0
                    else df_connections
                )

                series = row.series_x
                modality = row.modality_x
                folder = row.folder_x

                # For each component, this loop stores the CT and its connections
                temp = {
                    "study": study,
                    ct: {"modality": modality, "folder": folder},
                }

                # For saving the components in a format easier for the main pipeline
                folder_save = {
                    "study": study,
                    "patient_ID": row.patient_ID_x,
                    f"series_{modality}": series,
                    f"folder_{modality}": folder,
                }

                # This loop stores connection of the CT
                for _, row_y in df_connections.iterrows():
                    series_y = row_y.series_y
                    folder_y = row_y.folder_y
                    modality_y = row_y.modality_y

                    temp[row.series_y] = {
                        "modality": modality_y,
                        "folder": folder_y,
                        "conn_to": modality,
                    }

                    # Checks if there is already existing connection
                    key, key_series = self._check_save(
                        folder_save, modality_y, modality
                    )  # CT/MR
                    folder_save[key_series] = series_y
                    folder_save[key] = folder_y

                comp.append(temp)
                save_folder_comp.append(folder_save)

            # For rest of the edges left out, the connections are formed by going through the dictionary. For cases such as RTstruct-RTDose and PET-RTstruct
            rest_locs = df_temp.loc[
                ~df_temp.modality_x.isin(["CT", "MR"]),
                [
                    "series_x",
                    "modality_x",
                    "folder_x",
                    "series_y",
                    "modality_y",
                    "folder_y",
                ],
            ]
            for _, edge in rest_locs.iterrows():
                for k in range(len(ct_series)):
                    comp[k][edge["series_y"]] = {
                        "modality": edge["modality_y"],
                        "folder": edge["folder_y"],
                        "conn_to": edge["modality_x"],
                    }

                    # RTDOSE is connected via either RTstruct or/and CT, but we usually don't care, so naming it commonly
                    modality_origin = (
                        "CT"
                        if edge["modality_y"] == "RTDOSE"
                        else edge["modality_x"]
                    )

                    key, key_series = self._check_save(
                        save_folder_comp[k],
                        edge["modality_y"],
                        modality_origin,
                    )
                    save_folder_comp[k][key_series] = edge["series_y"]
                    save_folder_comp[k][key] = edge["folder_y"]

            remove_index = []
            if remove_less_comp:
                for j in range(len(ct_series)):
                    # Check if the number of nodes in a components isn't less than the query nodes, if yes then remove that component
                    present_modalities = set(
                        [
                            items.split("_")[1]
                            for items in save_folder_comp[j]
                            if items.split("_")[0] == "folder"
                        ]
                    )
                    # Checking if all the read modalities are present in a component
                    if desired_modalities.issubset(present_modalities):
                        remove_index.append(j)
                save_folder_comp = [
                    save_folder_comp[idx] for idx in remove_index
                ]
                comp = [comp[idx] for idx in remove_index]

            self.final_dict.extend(comp)
            final_df.extend(save_folder_comp)

        final_df = pd.DataFrame(final_df)
        return final_df

    def _check_save(
        self, save_dict: dict, node: str, dest: str
    ) -> Tuple[str, str]:
        key = f"folder_{node}_{dest}"
        key_series = f"series_{node}_{dest}"
        i = 1
        while key in save_dict:
            key = f"folder_{node}_{dest}_{i}"
            key_series = f"series_{node}_{dest}_{i}"
            i += 1
        return key, key_series
